import torch
import numpy as np
from PIL import Image
from typing import List, Union
from transformers import AutoModel, AutoTokenizer
import torchvision.transforms as T
from torchvision.transforms.functional import InterpolationMode
from decord import VideoReader, cpu

from .vqa_model import VQAScoreModel
from .fastchat_utils import get_conv_template

INTERNVL2_MODELS = {
    # InternVL2 Models
    'internvl2-1b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-1B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-1B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-2b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-2B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-2B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-4b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-4B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-4B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-8b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-8B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-8B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-26b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-26B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-26B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-40b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-40B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-40B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2-llama3-76b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2-Llama3-76B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2-Llama3-76B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    
    # InternVL2.5 Models
    'internvl2.5-1b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-1B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-1B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2.5-2b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-2B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-2B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2.5-4b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-4B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-4B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
        },
    },
    'internvl2.5-8b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-8B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-8B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
            'device_map': 'auto'
        },
    },
    'internvl2.5-26b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-26B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-26B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
            'device_map': 'auto'
        },
    },
    'internvl2.5-38b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-38B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-38B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
            'device_map': 'auto'
        },
    },
    'internvl2.5-78b': {
        'tokenizer': {
            'path': 'OpenGVLab/InternVL2_5-78B',
            'trust_remote_code': True,
            'use_fast': False,
        },
        'model': {
            'pretrained_model_name_or_path': 'OpenGVLab/InternVL2_5-78B',
            'torch_dtype': torch.bfloat16,
            'low_cpu_mem_usage': True,
            'use_flash_attn': True,
            'trust_remote_code': True,
            'device_map': 'auto'
        },
    },

    'internvl3-8b': {
    "tokenizer": {
        "path": "OpenGVLab/InternVL3-8B",
        "trust_remote_code": True,
        "use_fast": False,
    },
    "model": {
        "pretrained_model_name_or_path": "OpenGVLab/InternVL3-8B",
        "torch_dtype": torch.bfloat16,
        "low_cpu_mem_usage": True,
        "use_flash_attn": True,
        "trust_remote_code": True, 
        'device_map': 'auto'
        },
    },

    'internvl3-14b': {
    "tokenizer": {
        "path": "OpenGVLab/InternVL3-14B",
        "trust_remote_code": True,
        "use_fast": False,
    },
    "model": {
        "pretrained_model_name_or_path": "OpenGVLab/InternVL3-14B",
        "torch_dtype": torch.bfloat16,
        "low_cpu_mem_usage": True,
        "use_flash_attn": True,
        "trust_remote_code": True,
        'device_map': 'auto'
        },
    },

    'internvl3-78b': {
    "tokenizer": {
        "path": "OpenGVLab/InternVL3-78B",
        "trust_remote_code": True,
        "use_fast": False,
    },
    "model": {
        "pretrained_model_name_or_path": "OpenGVLab/InternVL3-78B",
        "torch_dtype": torch.bfloat16,
        "low_cpu_mem_usage": True,
        "use_flash_attn": True,
        "trust_remote_code": True,
        'device_map': 'auto'
        },
    }
}

IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)

class InternVL2Model(VQAScoreModel):
    video_mode = "direct"
    allows_image = True
    def __init__(self,
                 model_name='internvl2-8b',
                 device='cuda',
                 cache_dir=None):
        assert model_name in INTERNVL2_MODELS, f"Model {model_name} not found in INTERNVL2_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = INTERNVL2_MODELS[model_name]
        self.load_model()
        

    def load_model(self):
        tokenizer_path = self.model_info['tokenizer']['path']
        # self.model_info['model']['max_memory'] = {
        #         0: "45GB",  # GPU 0
        #         1: "45GB",  # GPU 1
        #         2: "45GB",  # GPU 4
        # }
        
        self.model = AutoModel.from_pretrained(
            **self.model_info['model']
        ).eval()#.to(self.device)
        
        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_path,
            **self.model_info['tokenizer']
        )
        
        # if 'internvl2' in self.model_name:
        self.device = next(self.model.parameters()).device # If there are multiple GPUs put the model on the first parameters GPU
        # elif 'internvl3' in self.model_name:
        #     self.device
    def build_transform(self, input_size=448):
        transform = T.Compose([
            T.Lambda(lambda img: img.convert('RGB') if img.mode != 'RGB' else img),
            T.Resize((input_size, input_size), interpolation=InterpolationMode.BICUBIC),
            T.ToTensor(),
            T.Normalize(mean=IMAGENET_MEAN, std=IMAGENET_STD)
        ])
        return transform

    def find_closest_aspect_ratio(self, aspect_ratio, target_ratios, width, height, image_size):
        best_ratio_diff = float('inf')
        best_ratio = (1, 1)
        area = width * height
        for ratio in target_ratios:
            target_aspect_ratio = ratio[0] / ratio[1]
            ratio_diff = abs(aspect_ratio - target_aspect_ratio)
            if ratio_diff < best_ratio_diff:
                best_ratio_diff = ratio_diff
                best_ratio = ratio
            elif ratio_diff == best_ratio_diff:
                if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                    best_ratio = ratio
        return best_ratio

    def dynamic_preprocess(self, image, min_num=1, max_num=12, image_size=448, use_thumbnail=False):
        orig_width, orig_height = image.size
        aspect_ratio = orig_width / orig_height

        target_ratios = set(
            (i, j) for n in range(min_num, max_num + 1) for i in range(1, n + 1) for j in range(1, n + 1) if
            i * j <= max_num and i * j >= min_num)
        target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

        target_aspect_ratio = self.find_closest_aspect_ratio(
            aspect_ratio, target_ratios, orig_width, orig_height, image_size)

        target_width = image_size * target_aspect_ratio[0]
        target_height = image_size * target_aspect_ratio[1]
        blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

        resized_img = image.resize((target_width, target_height))
        processed_images = []
        for i in range(blocks):
            box = (
                (i % (target_width // image_size)) * image_size,
                (i // (target_width // image_size)) * image_size,
                ((i % (target_width // image_size)) + 1) * image_size,
                ((i // (target_width // image_size)) + 1) * image_size
            )
            split_img = resized_img.crop(box)
            processed_images.append(split_img)
        
        if use_thumbnail and len(processed_images) != 1:
            thumbnail_img = image.resize((image_size, image_size))
            processed_images.append(thumbnail_img)
        return processed_images

    def load_image(self, image_file, input_size=448, max_num=12):
        if image_file.lower().endswith('.npy'):
            np_array = np.load(image_file)
            if np_array.ndim == 3:
                image = Image.fromarray(np_array.astype('uint8'), 'RGB')
            else:
                raise ValueError(f"Unexpected shape for NumPy array in {image_file}")
        else:
            image = Image.open(image_file).convert('RGB')
        
        transform = self.build_transform(input_size=input_size)
        images = self.dynamic_preprocess(image, image_size=input_size, use_thumbnail=True, max_num=max_num)
        pixel_values = [transform(image) for image in images]
        pixel_values = torch.stack(pixel_values)
        return pixel_values

    def get_video_index(self, bound, fps, max_frame, first_idx=0, num_segments=32):
        if bound:
            start, end = bound[0], bound[1]
        else:
            start, end = -100000, 100000
        start_idx = max(first_idx, round(start * fps))
        end_idx = min(round(end * fps), max_frame)
        seg_size = float(end_idx - start_idx) / num_segments
        frame_indices = np.array([
            int(start_idx + (seg_size / 2) + np.round(seg_size * idx))
            for idx in range(num_segments)
        ])
        return frame_indices

    def load_video(self, video_path, bound=None, input_size=448, max_num=1, num_segments=32):
        vr = VideoReader(video_path, ctx=cpu(0), num_threads=1)
        max_frame = len(vr) - 1
        fps = float(vr.get_avg_fps())

        pixel_values_list, num_patches_list = [], []
        transform = self.build_transform(input_size=input_size)
        frame_indices = self.get_video_index(bound, fps, max_frame, first_idx=0, num_segments=num_segments)
        for frame_index in frame_indices:
            img = Image.fromarray(vr[frame_index].numpy()).convert('RGB')
            img = self.dynamic_preprocess(img, image_size=input_size, use_thumbnail=True, max_num=max_num)
            pixel_values = [transform(tile) for tile in img]
            pixel_values = torch.stack(pixel_values)
            num_patches_list.append(pixel_values.shape[0])
            pixel_values_list.append(pixel_values)
        pixel_values = torch.cat(pixel_values_list)
        return pixel_values, num_patches_list

    def load_images(self, paths: List[str], num_frames: int = 16) -> List[Union[torch.Tensor, List[torch.Tensor]]]:
        processed_data = []
        num_patches_list = []
        for path in paths:
            if path.startswith(("http://", "https://")):
                raise NotImplementedError("Web link image/video inputs are not yet supported for this model. Please use a local path, or otherwise, make a Github issue request if this feature is necessary.")
            if path.lower().endswith(('.mp4', '.avi', '.mov', '.mkv')):  # Video file
                video_frames, video_num_patches = self.load_video(path, num_segments=num_frames)
                processed_data.append(video_frames)
                num_patches_list.append(video_num_patches) # You want a list of lists
            else:  # Image file or .npy file
                image_tensor = self.load_image(path)
                processed_data.append(image_tensor)
                num_patches_list.append(image_tensor.shape[0])
        return processed_data, num_patches_list

    def forward(self,
                paths: List[str],
                texts: List[str],
                num_frames: int=16,
                question_template: str = "Does this figure show \"{}\"? Please answer Yes or No.",
                answer_template: str = "Yes") -> torch.Tensor:
        assert len(paths) == len(texts), "Number of paths and texts must match"

        questions = [question_template.format(text) for text in texts]
        answers = [answer_template.format(text) for text in texts]
        processed_data, num_patches_list = self.load_images(paths, num_frames)

        lm_probs = []
        for data, question, num_patches, answer in zip(processed_data, questions, num_patches_list, answers):
            if isinstance(num_patches, list):  # Video
                video_prefix = ''.join([f'Frame{i+1}: <image>\n' for i in range(len(num_patches))])
                prompt = video_prefix + question
            else:  # Image
                prompt = '<image>\n' + question
                num_patches = [num_patches]

            # Reimplement chat method to allow for outputting scores :(
            # inputs = self.tokenizer(prompt, return_tensors="pt").to(self.device)
            # pixel_values = data.to(self.device).to(self.model.dtype)

            # with torch.no_grad():
            #     outputs = self.model.generate(
            #         **inputs,
            #         images=pixel_values,
            #         num_patches_list=[num_patches],
            #         max_new_tokens=1,
            #         do_sample=False,
            #         output_scores=True,
            #         return_dict_in_generate=True,
            #     )

            # Chat Arguments:
            pixel_values = data.to(self.device)
            pixel_values = pixel_values.to(self.model.dtype)
            generation_config = dict(max_new_tokens=1, do_sample=False, output_scores=True, return_dict_in_generate=True)
            IMG_START_TOKEN='<img>'
            IMG_END_TOKEN='</img>' 
            IMG_CONTEXT_TOKEN='<IMG_CONTEXT>'
            verbose=False
            history=None

            
            assert pixel_values is None or len(pixel_values) == sum(num_patches)
            img_context_token_id = self.tokenizer.convert_tokens_to_ids(IMG_CONTEXT_TOKEN)
            self.model.img_context_token_id = img_context_token_id

            template = get_conv_template(self.model.template)
            template.system_message = self.model.system_message
            eos_token_id = self.tokenizer.convert_tokens_to_ids(template.sep)

            history = [] if history is None else history
            for (old_question, old_answer) in history:
                template.append_message(template.roles[0], old_question)
                template.append_message(template.roles[1], old_answer)
            template.append_message(template.roles[0], prompt)
            template.append_message(template.roles[1], None)
            query = template.get_prompt()

            if verbose and pixel_values is not None:
                image_bs = pixel_values.shape[0]
                print(f'dynamic ViT batch size: {image_bs}')

            for patch in num_patches:
                image_tokens = IMG_START_TOKEN + IMG_CONTEXT_TOKEN * self.model.num_image_token * patch + IMG_END_TOKEN
                query = query.replace('<image>', image_tokens, 1)

            model_inputs = self.tokenizer(query, return_tensors='pt')
            input_ids = model_inputs['input_ids'].to(self.device)
            attention_mask = model_inputs['attention_mask'].to(self.device)
            generation_config['eos_token_id'] = eos_token_id
            with torch.amp.autocast(device_type='cuda', dtype=torch.bfloat16):
                outputs = self.model.generate(
                    pixel_values=pixel_values,
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    **generation_config
                )


            scores = outputs.scores[0]

            probs = torch.nn.functional.softmax(scores, dim=-1)
            yes_token_id = self.tokenizer.encode(answer)[0] # Option 1

            # Option 2:
            # if answer == "Yes":
            #     yes_token_id = self.tokenizer.encode(" " + answer, add_special_tokens=False)[0]

            lm_prob = probs[0, yes_token_id].item()
            lm_probs.append(lm_prob)

        return torch.tensor(lm_probs)
    
    def generate(self,
                images: List[str],
                texts: List[str],
                num_frames: int=16,
                max_new_tokens: int = 256) -> torch.Tensor:
        assert len(images) == len(texts), "Number of paths and texts must match"

        questions = texts
        processed_data, num_patches_list = self.load_images(images, num_frames)

        gen_outputs = []
        for data, question, num_patches in zip(processed_data, questions, num_patches_list):
            if isinstance(num_patches, list):  # Video
                video_prefix = ''.join([f'Frame{i+1}: <image>\n' for i in range(len(num_patches))])
                prompt = video_prefix + question
            else:  # Image
                prompt = '<image>\n' + question
                num_patches = [num_patches]

            pixel_values = data.to(self.device)
            generation_config = dict(max_new_tokens=max_new_tokens, do_sample=False, output_scores=True, return_dict_in_generate=True)
            IMG_START_TOKEN='<img>'
            IMG_END_TOKEN='</img>' 
            IMG_CONTEXT_TOKEN='<IMG_CONTEXT>'
            verbose=False
            history=None

            
            assert pixel_values is None or len(pixel_values) == sum(num_patches)
            img_context_token_id = self.tokenizer.convert_tokens_to_ids(IMG_CONTEXT_TOKEN)
            self.model.img_context_token_id = img_context_token_id

            template = get_conv_template(self.model.template)
            template.system_message = self.model.system_message
            eos_token_id = self.tokenizer.convert_tokens_to_ids(template.sep.strip())

            history = [] if history is None else history
            for (old_question, old_answer) in history:
                template.append_message(template.roles[0], old_question)
                template.append_message(template.roles[1], old_answer)
            template.append_message(template.roles[0], prompt)
            template.append_message(template.roles[1], None)
            query = template.get_prompt()

            if verbose and pixel_values is not None:
                image_bs = pixel_values.shape[0]
                print(f'dynamic ViT batch size: {image_bs}')

            for patch in num_patches:
                image_tokens = IMG_START_TOKEN + IMG_CONTEXT_TOKEN * self.model.num_image_token * patch + IMG_END_TOKEN
                query = query.replace('<image>', image_tokens, 1)

            model_inputs = self.tokenizer(query, return_tensors='pt')
            input_ids = model_inputs['input_ids'].to(self.model.device)
            attention_mask = model_inputs['attention_mask'].to(self.device)
            generation_config['eos_token_id'] = eos_token_id
            with torch.amp.autocast(device_type='cuda', dtype=torch.bfloat16):
                outputs = self.model.generate(
                    pixel_values=pixel_values,
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    **generation_config
                )


            outputs = self.tokenizer.decode(outputs.sequences[0], skip_special_tokens=True).strip()
            gen_outputs.append(outputs)

        return gen_outputs



# Copyright (c) Meta Platforms, Inc. and affiliates.
import math

import torch
import torch.nn.functional as F
from torch import nn

from ...core.utils import get_init_fn
from ...core.vision_projector.base import BaseProjector


class AdaptiveAvgPooling(nn.Module):
    def __init__(self, pooling_ratio=2):
        super(AdaptiveAvgPooling, self).__init__()
        self.pooling_ratio = pooling_ratio

    def forward(self, x):
        b, num_tokens, c = x.shape
        h = int(math.sqrt(num_tokens))
        assert h * h == num_tokens

        shape = (h // self.pooling_ratio, h // self.pooling_ratio)
        x = x.permute(0, 2, 1).reshape(b, -1, h, h)
        x = F.adaptive_avg_pool2d(x, shape)
        x = x.flatten(2).transpose(1, 2)

        return x


class MLPProjector(BaseProjector):
    def __init__(self, args):
        super().__init__()
        self.setup_projector(args)
        self.pooling_ratio = args.pooling_ratio
        self.adaptive_avg_pool = AdaptiveAvgPooling(pooling_ratio=args.pooling_ratio)
        self.remove_vision_class_token = args.remove_vision_class_token

    def init_tensors(self):
        self.init_method(self.projector[0].weight)
        self.init_method(self.projector[0].bias)
        self.init_method(self.projector[2].weight)
        self.init_method(self.projector[2].bias)

    def setup_projector(self, args):
        self.init_method = get_init_fn(args.mlp_init, args.dim, init_depth=None)
        input_size = args.vision_model["width"]
        output_size = args.dim
        self.projector = nn.Sequential(
            nn.Linear(
                in_features=input_size,
                out_features=output_size,
                bias=True,
                dtype=torch.get_default_dtype(),
            ),
            nn.GELU(),
            nn.Linear(
                in_features=output_size,
                out_features=output_size,
                bias=True,
                dtype=torch.get_default_dtype(),
            ),
        )



"""Tests for avqt module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_avqt_basics():
    from ayase.modules.avqt import AVQTModule
    _test_module_basics(AVQTModule, "avqt")

def test_avqt_no_reference(image_sample):
    from ayase.modules.avqt import AVQTModule
    m = AVQTModule()
    result = m.process(image_sample)
    assert result is image_sample



from collections import OrderedDict
from itertools import repeat
import collections.abc
import math

import torch
import torch.nn.functional as F
from torch import nn

from fairscale.nn.checkpoint.checkpoint_activations import checkpoint_wrapper

from ..models.eva_vit import convert_weights_to_fp16
from ..common.dist_utils import download_cached_file

class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1):
        super().__init__()

        # all conv layers have stride 1. an avgpool is performed after the second convolution when stride > 1
        self.conv1 = nn.Conv2d(inplanes, planes, 1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu1 = nn.ReLU(inplace=True)

        self.conv2 = nn.Conv2d(planes, planes, 3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.relu2 = nn.ReLU(inplace=True)

        self.avgpool = nn.AvgPool2d(stride) if stride > 1 else nn.Identity()

        self.conv3 = nn.Conv2d(planes, planes * self.expansion, 1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.relu3 = nn.ReLU(inplace=True)

        self.downsample = None
        self.stride = stride

        if stride > 1 or inplanes != planes * Bottleneck.expansion:
            # downsampling layer is prepended with an avgpool, and the subsequent convolution has stride 1
            self.downsample = nn.Sequential(OrderedDict([
                ("-1", nn.AvgPool2d(stride)),
                ("0", nn.Conv2d(inplanes, planes * self.expansion, 1, stride=1, bias=False)),
                ("1", nn.BatchNorm2d(planes * self.expansion))
            ]))

    def forward(self, x: torch.Tensor):
        identity = x

        out = self.relu1(self.bn1(self.conv1(x)))
        out = self.relu2(self.bn2(self.conv2(out)))
        out = self.avgpool(out)
        out = self.bn3(self.conv3(out))

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu3(out)
        return out


class AttentionPool2d(nn.Module):
    def __init__(self, spacial_dim: int, embed_dim: int, num_heads: int, output_dim: int = None):
        super().__init__()
        self.positional_embedding = nn.Parameter(torch.randn(spacial_dim ** 2 + 1, embed_dim) / embed_dim ** 0.5)
        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.c_proj = nn.Linear(embed_dim, output_dim or embed_dim)
        self.num_heads = num_heads

    def forward(self, x):
        x = x.reshape(x.shape[0], x.shape[1], x.shape[2] * x.shape[3]).permute(2, 0, 1)  # NCHW -> (HW)NC
        x = torch.cat([x.mean(dim=0, keepdim=True), x], dim=0)  # (HW+1)NC
        x = x + self.positional_embedding[:, None, :].to(x.dtype)  # (HW+1)NC
        x, _ = F.multi_head_attention_forward(
            query=x, key=x, value=x,
            embed_dim_to_check=x.shape[-1],
            num_heads=self.num_heads,
            q_proj_weight=self.q_proj.weight,
            k_proj_weight=self.k_proj.weight,
            v_proj_weight=self.v_proj.weight,
            in_proj_weight=None,
            in_proj_bias=torch.cat([self.q_proj.bias, self.k_proj.bias, self.v_proj.bias]),
            bias_k=None,
            bias_v=None,
            add_zero_attn=False,
            dropout_p=0,
            out_proj_weight=self.c_proj.weight,
            out_proj_bias=self.c_proj.bias,
            use_separate_proj_weight=True,
            training=self.training,
            need_weights=False
        )

        return x[0]
    

class LayerNorm(nn.LayerNorm):
    """Subclass torch's LayerNorm to handle fp16."""

    def forward(self, x: torch.Tensor):
        orig_type = x.dtype
        ret = super().forward(x.type(torch.float32))
        return ret.type(orig_type)


class QuickGELU(nn.Module):
    def forward(self, x: torch.Tensor):
        return x * torch.sigmoid(1.702 * x)


class ResidualAttentionBlock(nn.Module):
    def __init__(self, d_model: int, n_head: int, attn_mask: torch.Tensor = None, use_grad_checkpointing=False):
        super().__init__()

        self.attn = nn.MultiheadAttention(d_model, n_head)
        self.ln_1 = LayerNorm(d_model)
        self.mlp = nn.Sequential(OrderedDict([
            ("c_fc", nn.Linear(d_model, d_model * 4)),
            ("gelu", QuickGELU()),
            ("c_proj", nn.Linear(d_model * 4, d_model))
        ]))
        self.ln_2 = LayerNorm(d_model)
        self.attn_mask = attn_mask

        if use_grad_checkpointing:
            self.attn = checkpoint_wrapper(self.attn)
            self.mlp = checkpoint_wrapper(self.mlp)
            
    def attention(self, x: torch.Tensor):
        self.attn_mask = self.attn_mask.to(dtype=x.dtype, device=x.device) if self.attn_mask is not None else None
        return self.attn(x, x, x, need_weights=False, attn_mask=self.attn_mask)[0]

    def forward(self, x: torch.Tensor):
        x = x + self.attention(self.ln_1(x))
        x = x + self.mlp(self.ln_2(x))
        return x


class Transformer(nn.Module):
    def __init__(self, width: int, layers: int, heads: int, attn_mask: torch.Tensor = None, use_grad_checkpointing=False):
        super().__init__()
        self.width = width
        self.layers = layers
        self.resblocks = nn.Sequential(*[ResidualAttentionBlock(width, heads, attn_mask, use_grad_checkpointing and i>12) for i in range(layers)])

    def forward(self, x: torch.Tensor):
        return self.resblocks(x)


class VisionTransformer(nn.Module):
    def __init__(self, input_resolution: int, patch_size: int, width: int, layers: int, heads: int, use_grad_checkpointing: bool):
        super().__init__()
        self.input_resolution = input_resolution
        self.num_features = width
        self.num_heads = heads
        self.num_patches = (input_resolution // patch_size) ** 2
        self.conv1 = nn.Conv2d(in_channels=3, out_channels=width, kernel_size=patch_size, stride=patch_size, bias=False)

        scale = width ** -0.5
        self.class_embedding = nn.Parameter(scale * torch.randn(width))
        self.positional_embedding = nn.Parameter(scale * torch.randn(self.num_patches + 1, width))
        self.ln_pre = LayerNorm(width)
        
        self.transformer = Transformer(width, layers-1, heads, use_grad_checkpointing=use_grad_checkpointing)
           
#         self.ln_final = LayerNorm(width)

    def forward(self, x: torch.Tensor):

        x = self.conv1(x)  # shape = [*, width, grid, grid]
        x = x.reshape(x.shape[0], x.shape[1], -1)  # shape = [*, width, grid ** 2]
        x = x.permute(0, 2, 1)  # shape = [*, grid ** 2, width]
        x = torch.cat([self.class_embedding.to(x.dtype) + torch.zeros(x.shape[0], 1, x.shape[-1], dtype=x.dtype, device=x.device), x], dim=1)  # shape = [*, grid ** 2 + 1, width]
        x = x + self.positional_embedding.to(x.dtype)
        x = self.ln_pre(x)

        x = x.permute(1, 0, 2)  # NLD -> LND
        x = self.transformer(x)
        x = x.permute(1, 0, 2)  # LND -> NLD
        
#         x = self.ln_final(x)
        return x
    
    
            
# From PyTorch internals
def _ntuple(n):
    def parse(x):
        if isinstance(x, collections.abc.Iterable):
            return x
        return tuple(repeat(x, n))
    return parse
to_2tuple = _ntuple(2)        
def interpolate_pos_embed(model, state_dict, interpolation: str = 'bicubic', seq_dim=1):
    # Rescale the grid of position embeddings when loading from state_dict
    old_pos_embed = state_dict.get('positional_embedding', None)
    
    grid_size = round((model.positional_embedding.shape[0] - 1) ** 0.5)
    if old_pos_embed is None:
        return
    grid_size = to_2tuple(grid_size)
    extra_tokens = 1  # FIXME detect different token configs (ie no class token, or more)
    new_seq_len = grid_size[0] * grid_size[1] + extra_tokens
    if new_seq_len == old_pos_embed.shape[0]:
        return

    if extra_tokens:
        pos_emb_tok, pos_emb_img = old_pos_embed[:extra_tokens], old_pos_embed[extra_tokens:]
    else:
        pos_emb_tok, pos_emb_img = None, old_pos_embed
        
    old_grid_size = to_2tuple(int(math.sqrt(len(pos_emb_img))))

    print('Resizing position embedding grid-size from %s to %s', old_grid_size, grid_size)
    pos_emb_img = pos_emb_img.reshape(1, old_grid_size[0], old_grid_size[1], -1).permute(0, 3, 1, 2)
    pos_emb_img = F.interpolate(
        pos_emb_img,
        size=grid_size,
        mode=interpolation,
        align_corners=True,
    )
    pos_emb_img = pos_emb_img.permute(0, 2, 3, 1).reshape(1, grid_size[0] * grid_size[1], -1)[0]
    if pos_emb_tok is not None:
        new_pos_embed = torch.cat([pos_emb_tok, pos_emb_img], dim=0)
    else:
        new_pos_embed = pos_emb_img
    state_dict['positional_embedding'] = new_pos_embed
    
    
def create_clip_vit_L(img_size=224,use_checkpoint=False,precision="fp16"):
    model = VisionTransformer(
            input_resolution=img_size,
            patch_size=14,
            width=1024,
            layers=22,
            heads=16,
            use_grad_checkpointing=use_checkpoint,
        )         
    url = "https://storage.googleapis.com/sfr-vision-language-research/LAVIS/models/BLIP2/clip_vit_L.pth"
    cached_file = download_cached_file(
        url, check_hash=False, progress=True
    )
    state_dict = torch.load(cached_file, map_location="cpu")    
    interpolate_pos_embed(model,state_dict)
    
    incompatible_keys = model.load_state_dict(state_dict, strict=False)
    # print(incompatible_keys)
    
    if precision == "fp16":
        convert_weights_to_fp16(model)
    return model


"""Professional Production Quality metrics module.

Assesses cinematography-level quality attributes:

  color_grading_score  — colour consistency across the clip (0-100)
  white_balance_score  — accuracy / neutrality of white balance (0-100)
  exposure_consistency — stability of exposure over time (0-100)
  focus_quality        — sharpness / depth-of-field quality (0-100)
  banding_severity     — colour banding in gradients (0-100, lower=better)

All metrics are pure OpenCV + NumPy (no ML).  Only videos are
processed (images get a single-frame assessment where applicable).
"""

import logging
from typing import List

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class ProductionQualityModule(PipelineModule):
    name = "production_quality"
    description = "Professional production quality (colour, exposure, focus, banding)"
    default_config = {
        "max_frames": 150,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.max_frames = self.config.get("max_frames", 150)

    # ------------------------------------------------------------------
    # Individual metrics
    # ------------------------------------------------------------------

    @staticmethod
    def _color_grading_score(frames: List[np.ndarray]) -> float:
        """How consistent is the colour palette across the clip?

        We measure the standard deviation of per-frame mean colour
        in LAB space.  Lower deviation → higher score.
        """
        if len(frames) < 2:
            return 100.0

        means = []
        for f in frames:
            lab = cv2.cvtColor(f, cv2.COLOR_BGR2LAB).astype(np.float32)
            means.append(lab.mean(axis=(0, 1)))

        means = np.stack(means)
        dev = float(means.std(axis=0).mean())
        # Normalise: dev ~0 → 100, dev ~20+ → 0
        return float(np.clip(100.0 - dev * 5.0, 0, 100))

    @staticmethod
    def _white_balance_score(frame: np.ndarray) -> float:
        """Grey-world assumption: how close is the mean colour to neutral grey?

        Neutral means R≈G≈B.  The further away, the stronger the
        colour cast (worse WB).
        """
        bgr_mean = frame.mean(axis=(0, 1))  # [B, G, R]
        overall = bgr_mean.mean()
        if overall < 1:
            return 50.0

        deviation = np.std(bgr_mean / overall)
        # deviation ~0 → perfect, ~0.3+ → bad
        return float(np.clip(100.0 - deviation * 300.0, 0, 100))

    @staticmethod
    def _exposure_consistency(frames: List[np.ndarray]) -> float:
        """Measures stability of luminance across frames."""
        if len(frames) < 2:
            return 100.0

        lums = []
        for f in frames:
            gray = cv2.cvtColor(f, cv2.COLOR_BGR2GRAY)
            lums.append(float(gray.mean()))

        cv_val = float(np.std(lums) / (np.mean(lums) + 1e-6))
        # CV ~0 → 100 (stable), CV ~0.2+ → 0 (unstable)
        return float(np.clip(100.0 - cv_val * 500.0, 0, 100))

    @staticmethod
    def _focus_quality(frame: np.ndarray) -> float:
        """Sharpness via Laplacian variance, mapped to 0-100."""
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        lap_var = cv2.Laplacian(gray, cv2.CV_64F).var()
        # lap_var ~1000+ is very sharp, ~10 is very blurry
        return float(np.clip(lap_var / 10.0, 0, 100))

    @staticmethod
    def _banding_severity(frame: np.ndarray) -> float:
        """Detects colour banding in smooth gradients.

        Method: look at the Y channel, compute horizontal and
        vertical second-order differences.  In smooth regions
        banding creates periodic step-like patterns which have high
        second-derivative amplitude on an otherwise flat region.
        """
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float32)

        # Second-order horizontal difference
        d2 = np.abs(gray[:, 2:] - 2 * gray[:, 1:-1] + gray[:, :-2])

        # Focus on smooth regions: where local gradient is small
        grad = np.abs(gray[:, 1:] - gray[:, :-1])
        # Pad grad to match d2 shape
        grad = grad[:, :-1]
        smooth_mask = grad < 5.0  # smooth region

        if smooth_mask.sum() < 100:
            return 0.0

        # Banding score = mean of second derivative in smooth regions
        banding = float(d2[smooth_mask].mean())
        # Scale: banding ~0 → no issue, ~5+ → severe
        return float(np.clip(banding * 20.0, 0, 100))

    # ------------------------------------------------------------------
    # Process
    # ------------------------------------------------------------------

    def process(self, sample: Sample) -> Sample:
        try:
            if sample.is_video:
                return self._process_video(sample)
            else:
                return self._process_image(sample)
        except Exception as e:
            logger.error(f"Production quality failed for {sample.path}: {e}")
            return sample

    def _process_image(self, sample: Sample) -> Sample:
        img = cv2.imread(str(sample.path))
        if img is None:
            return sample

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        sample.quality_metrics.white_balance_score = self._white_balance_score(img)
        sample.quality_metrics.focus_quality = self._focus_quality(img)
        sample.quality_metrics.banding_severity = self._banding_severity(img)

        return sample

    def _process_video(self, sample: Sample) -> Sample:
        cap = cv2.VideoCapture(str(sample.path))
        if not cap.isOpened():
            return sample

        total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

        # Determine which frame indices to keep (uniform subsample if needed)
        if total_frames > 0 and total_frames > self.max_frames:
            keep_indices = set(np.linspace(0, total_frames - 1, self.max_frames, dtype=int))
        else:
            keep_indices = None  # keep all

        frames = []
        wb_scores = []
        focus_scores = []
        banding_scores = []
        frame_idx = 0
        kept_count = 0

        while True:
            ret, frame = cap.read()
            if not ret:
                break

            if keep_indices is None or frame_idx in keep_indices:
                # Keep every sampled frame for temporal metrics (colour/exposure)
                frames.append(frame)
                # Subsample spatial metrics (every 5th kept frame)
                if kept_count % 5 == 0:
                    wb_scores.append(self._white_balance_score(frame))
                    focus_scores.append(self._focus_quality(frame))
                    banding_scores.append(self._banding_severity(frame))
                kept_count += 1

            frame_idx += 1

        cap.release()

        if not frames:
            return sample

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        sample.quality_metrics.color_grading_score = self._color_grading_score(frames)
        sample.quality_metrics.exposure_consistency = self._exposure_consistency(frames)
        sample.quality_metrics.white_balance_score = (
            float(np.mean(wb_scores)) if wb_scores else None
        )
        sample.quality_metrics.focus_quality = (
            float(np.mean(focus_scores)) if focus_scores else None
        )
        sample.quality_metrics.banding_severity = (
            float(np.mean(banding_scores)) if banding_scores else None
        )

        logger.debug(
            f"Production for {sample.path.name}: "
            f"cg={sample.quality_metrics.color_grading_score} "
            f"wb={sample.quality_metrics.white_balance_score} "
            f"exp={sample.quality_metrics.exposure_consistency} "
            f"foc={sample.quality_metrics.focus_quality} "
            f"band={sample.quality_metrics.banding_severity or 0.0:.0f}"
        )

        return sample



import torch
import torch.nn as nn
import torch.nn.functional as F
from ayase.third_party.rife.warplayer import warp

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),        
        nn.PReLU(out_planes)
    )

def conv_bn(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=False),
        nn.BatchNorm2d(out_planes),
        nn.PReLU(out_planes)
    )

class IFBlock(nn.Module):
    def __init__(self, in_planes, c=64):
        super(IFBlock, self).__init__()
        self.conv0 = nn.Sequential(
            conv(in_planes, c//2, 3, 2, 1),
            conv(c//2, c, 3, 2, 1),
            )
        self.convblock0 = nn.Sequential(
            conv(c, c),
            conv(c, c)
        )
        self.convblock1 = nn.Sequential(
            conv(c, c),
            conv(c, c)
        )
        self.convblock2 = nn.Sequential(
            conv(c, c),
            conv(c, c)
        )
        self.convblock3 = nn.Sequential(
            conv(c, c),
            conv(c, c)
        )
        self.conv1 = nn.Sequential(
            nn.ConvTranspose2d(c, c//2, 4, 2, 1),
            nn.PReLU(c//2),
            nn.ConvTranspose2d(c//2, 4, 4, 2, 1),
        )
        self.conv2 = nn.Sequential(
            nn.ConvTranspose2d(c, c//2, 4, 2, 1),
            nn.PReLU(c//2),
            nn.ConvTranspose2d(c//2, 1, 4, 2, 1),
        )

    def forward(self, x, flow, scale=1):
        x = F.interpolate(x, scale_factor= 1. / scale, mode="bilinear", align_corners=False, recompute_scale_factor=False)
        flow = F.interpolate(flow, scale_factor= 1. / scale, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 1. / scale
        feat = self.conv0(torch.cat((x, flow), 1))
        feat = self.convblock0(feat) + feat
        feat = self.convblock1(feat) + feat
        feat = self.convblock2(feat) + feat
        feat = self.convblock3(feat) + feat        
        flow = self.conv1(feat)
        mask = self.conv2(feat)
        flow = F.interpolate(flow, scale_factor=scale, mode="bilinear", align_corners=False, recompute_scale_factor=False) * scale
        mask = F.interpolate(mask, scale_factor=scale, mode="bilinear", align_corners=False, recompute_scale_factor=False)
        return flow, mask
        
class IFNet(nn.Module):
    def __init__(self):
        super(IFNet, self).__init__()
        self.block0 = IFBlock(7+4, c=90)
        self.block1 = IFBlock(7+4, c=90)
        self.block2 = IFBlock(7+4, c=90)
        self.block_tea = IFBlock(10+4, c=90)
        # self.contextnet = Contextnet()
        # self.unet = Unet()

    def forward(self, x, scale_list=[4, 2, 1], training=False):
        if training == False:
            channel = x.shape[1] // 2
            img0 = x[:, :channel]
            img1 = x[:, channel:]
        flow_list = []
        merged = []
        mask_list = []
        warped_img0 = img0
        warped_img1 = img1
        flow = (x[:, :4]).detach() * 0
        mask = (x[:, :1]).detach() * 0
        loss_cons = 0
        block = [self.block0, self.block1, self.block2]
        for i in range(3):
            f0, m0 = block[i](torch.cat((warped_img0[:, :3], warped_img1[:, :3], mask), 1), flow, scale=scale_list[i])
            f1, m1 = block[i](torch.cat((warped_img1[:, :3], warped_img0[:, :3], -mask), 1), torch.cat((flow[:, 2:4], flow[:, :2]), 1), scale=scale_list[i])
            flow = flow + (f0 + torch.cat((f1[:, 2:4], f1[:, :2]), 1)) / 2
            mask = mask + (m0 + (-m1)) / 2
            mask_list.append(mask)
            flow_list.append(flow)
            warped_img0 = warp(img0, flow[:, :2])
            warped_img1 = warp(img1, flow[:, 2:4])
            merged.append((warped_img0, warped_img1))
        '''
        c0 = self.contextnet(img0, flow[:, :2])
        c1 = self.contextnet(img1, flow[:, 2:4])
        tmp = self.unet(img0, img1, warped_img0, warped_img1, mask, flow, c0, c1)
        res = tmp[:, 1:4] * 2 - 1
        '''
        for i in range(3):
            mask_list[i] = torch.sigmoid(mask_list[i])
            merged[i] = merged[i][0] * mask_list[i] + merged[i][1] * (1 - mask_list[i])
            # merged[i] = torch.clamp(merged[i] + res, 0, 1)        
        return flow_list, mask_list[2], merged



import os 
import sys
import torch
import numpy as np
from PIL import Image
from typing import List, Union
import copy
import warnings
import requests
import tempfile

from .tarsier.tasks.utils import load_model_and_processor
from .tarsier.dataset.custom_data_parsers.utils import put_pred_to_data_dict, get_prompt_from_data_dict
from .tarsier.dataset.utils import format_one_sample, get_visual_type
# sys.path = sys.path[2:]
# print(sys.path)
import yaml


from .vqa_model import VQAScoreModel

warnings.filterwarnings("ignore")

current_path = os.path.dirname(os.path.abspath(__file__))
TARSIER_MODELS = {
    'tarsier-recap-7b': {
        'model': {
            'path': "omni-research/Tarsier2-Recap-7b" ,
            'config': f'{current_path}/tarsier/configs/tarser2_default_config.yaml',
        },
    },

    'tarsier2-7b': {
        'model': {
            'path': "omni-research/Tarsier2-7b-0115" ,
            'config': f'{current_path}/tarsier/configs/tarser2_default_config.yaml',
        },
    },
}

class TarsierModel(VQAScoreModel):
    video_mode = "direct"
    allows_image = True
    def __init__(self,
                 model_name='tarsier-recap-7b',
                 device='cuda',
                 cache_dir=None):
        assert model_name in TARSIER_MODELS, f"Model {model_name} not found in TARSIER_MODELS"
        self.model_name = model_name
        self.device = device
        self.cache_dir = cache_dir
        self.model_info = TARSIER_MODELS[model_name]
        self.load_model()

        self.tmp_files = []

    def load_model(self):
        # sys.path = [os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'vqascore_models/tarsier')] + sys.path
        # sys.path = [os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'vqascore_models/tarsier/models')] + sys.path
        # print(f'sys.path {sys.path}')
        model_path = self.model_info['model']['path']
        config_path = self.model_info['model']['config']
        
        data_config = yaml.safe_load(open(config_path, 'r'))
        self.model, self.processor = load_model_and_processor(model_path, data_config=data_config)
        self.model.eval()
        # sys.path = sys.path[2:]

    def load_images(self, prompt, video_file, generate_kwargs):
        if video_file.startswith(("http://", "https://")):
            tmpfile = None
            try:
                # 1. Read video
                response = requests.get(video_file, stream=True)
                response.raise_for_status()

                # 2. Create tempfile to write  video to
                with tempfile.NamedTemporaryFile(delete=False, suffix='.mp4') as tmpfile:
                    self.tmp_files.append(tmpfile.name)

                    for chunk in response.iter_content(chunk_size=8192):
                        tmpfile.write(chunk)

                    
                    tmpfile.flush()
                # 3. Replace video file link with the temp file path name
                    
                    video_file=tmpfile.name
            except Exception as e:
                if tmpfile:
                    os.remove(tmpfile.name)
                    self.tmp_files.remove(tmpfile.name)

                
                print(f"Exception while downloading video {video_file} \n\nError: {e}")
                raise

        sample = format_one_sample(video_file, prompt)
        batch_data = self.processor(sample)
        
        model_inputs = {}
        for k, v in batch_data.items():
            if not isinstance(v, torch.Tensor):
                continue
            model_inputs[k] = v.to(self.model.device)
            
        outputs = self.model.generate(
            **model_inputs,
            **generate_kwargs,
        )
        
        output_text = self.processor.processor.tokenizer.decode(
            outputs[0][model_inputs['input_ids'][0].shape[0]:], 
            skip_special_tokens=True
        )
        return output_text, model_inputs
    def clean_temp_files(self):
        # Clean up tempfiles created:
        for file_path in self.tmp_files:
            try:
                os.remove(file_path)
            except OSError:
                pass
        self.tmp_files.clear()
    def forward(self,
                images: List[str],
                texts: List[str],
                num_frames: int=16,
                question_template: str = "Does this video show \"{}\"? Answer the question with Yes or No.",
                answer_template: str = "Yes") -> torch.Tensor:
        assert len(images) == len(texts), "Number of images and texts must match"
    
        lm_probs = []
        for path, text in zip(images, texts):
            if not get_visual_type(path) in ['video', 'gif']:
                print(f'This model supports only video inference. {path} is invalid')
                lm_probs.append(0.0)
                continue

            question = question_template.format(text)
            answer = answer_template.format(text)
            
            generate_kwargs = {
                "do_sample": False,
                "max_new_tokens": 1,
                "top_p": 1.0,
                "temperature": 0,
                "use_cache": True,
                "output_scores": True,
                "return_dict_in_generate": True,
            }
            
            _, model_inputs = self.load_images(question, path, generate_kwargs)
            
            outputs = self.model.generate(
                **model_inputs,
                **generate_kwargs,
            )
            
            scores = outputs.scores[0]
            probs = torch.nn.functional.softmax(scores, dim=-1)
            yes_token_id = self.processor.processor.tokenizer.encode(answer)[0]
            lm_prob = probs[0, yes_token_id].item()
            lm_probs.append(lm_prob)

        self.clean_temp_files()

        return torch.tensor(lm_probs)
    
    def generate(self,
                images: List[str],
                texts: List[str],
                num_frames: int = 4,
                max_new_tokens: int = 256) -> List[str]:
        assert len(images) == len(texts), "Number of paths and texts must match"
        
        generated_outputs = []
        for path, text in zip(images, texts):
            if not get_visual_type(path) in ['video', 'gif']:
                print(f'This model supports only video inference. {path} is invalid')
                generated_outputs.append("")
                continue
                
            generate_kwargs = {
                "do_sample": False,
                "max_new_tokens": max_new_tokens,
                "top_p": 1.0,
                "temperature": 0,
                "use_cache": True
            }
            
            output_text, _ = self.load_images(text, path, generate_kwargs)
            generated_outputs.append(output_text.strip())

        
        self.clean_temp_files()

        return generated_outputs


# Copyright (c) Meta Platforms, Inc. and affiliates.

import math
from functools import reduce
from logging import getLogger
from typing import Any, Callable, Tuple

import numpy as np
import torch
import torchvision.transforms as tv
from PIL import Image
from torchvision.transforms import functional as F
from torchvision.transforms.functional import InterpolationMode

logger = getLogger()


MEAN = (0.5, 0.5, 0.5)
STD = (0.5, 0.5, 0.5)


def get_image_transform(
    vision_input_type: str = "vanilla",
    image_res: int = 336,
    max_num_tiles: int = 1,
    normalize_img: bool = True,
) -> Tuple[Callable, int]:

    if vision_input_type == "thumb+tile":
        transforms = VariableSizeImageTransform(
            size=image_res,
            max_num_tiles=max_num_tiles,
            normalize_img=normalize_img,
            use_thumbnail="before",
        )
    else:
        transforms = ImageTransform(
            size=image_res,
            normalize_img=normalize_img,
        )

    logger.info(
        f"Initalized transforms with: vision_input_type: '{vision_input_type}' and max_num_tiles: {max_num_tiles}."
    )

    return transforms


class ImageTransform(object):
    """
    Image transform will resize the longer edge to a given size and pad the shorter edge with mean pixel value of the image.
    """

    def __init__(
        self,
        size: int = 336,
        normalize_img: bool = True,
    ) -> None:
        self.size = size
        self._mean = MEAN
        self._std = STD

        logger.info(f"ImageTransform size: {self.size}")

        self.to_tensor = tv.ToTensor()
        self.normalize = (
            tv.Normalize(
                mean=self._mean,
                std=self._std,
                inplace=True,
            )
            if normalize_img
            else lambda x: x
        )

    def __call__(self, image: Image.Image):
        w, h = image.size
        image = F.resize(
            image, (self.size, self.size), interpolation=InterpolationMode.BICUBIC
        )
        image = self.to_tensor(image)
        image = self.normalize(image)

        # Add chunk dim to make it compatible with existing dataloaders
        image = image.view(1, 3, self.size, self.size)
        return image, (w, h)

    def _transform_torch_tensor(self, image: torch.Tensor):
        h, w = image.shape[-2:]  # Image shape (C, H, W) or (N, C, H, W)
        image = F.resize(
            image, size=(self.size, self.size), interpolation=InterpolationMode.BICUBIC
        )
        image = (
            image.to(torch.float32) / 255.0
        )  # Convert to float and scale to [0, 1] range
        image = self.normalize(image)
        return image, (w, h)


class VariableSizeImageTransform(object):
    """
    The variable size image transform will resize the image dynamically
    based on the image aspect ratio and the number of image chunks we allow.

    The algorithm will not upsample low-res images to fit a certain aspect
    ratio, because that leads to a significant degradation in image quality.

    For example, if an input image is of size 300x800, and we want to allow
    a maximum of 16 image chunks, it will find the closest aspect ratio that
    is allowed within 16 image chunks, i.e., 2:5 = 2 horizontal patches and
    5 vertical patches, giving a total of 10 chunks.

    The image will then be resized to products of the base size (default is
    224px because MetaCLIP takes that), so in this case it will  be resized to
    2*224:5*224 = 448:1120, where we maintain the original aspect ratio and
    pad with the mean value for the rest. This approach minimizes the amount
    of padding required for any arbitrary resolution.

    The final output will therefore be of shape (11, 3, 224, 224), where 10
    patches are coming from the resizing and chunking, and the first patch
    is a downsampled version of the image that preserves aspect ratios.
    """

    def __init__(
        self,
        size: int = 336,
        normalize_img: bool = True,
        max_num_tiles: int = 1,
        use_thumbnail: str = "no",
        area_limit: bool = False,
    ) -> None:
        self.size = size
        self._mean = MEAN
        self._std = STD

        logger.info(f"VariableSizeImageTransform size: {self.size}")

        self.to_tensor = tv.ToTensor()
        self.normalize = (
            tv.Normalize(
                mean=self._mean,
                std=self._std,
                inplace=True,
            )
            if normalize_img
            else lambda x: x
        )
        self.area_limit = area_limit
        self.max_num_tiles = max_num_tiles
        self.use_thumbnail = use_thumbnail
        if self.use_thumbnail != "no":
            self.thumbnail_transform = ImageTransform(
                size=self.size,
                normalize_img=normalize_img,
            )

    @staticmethod
    def _factors(n: int):
        """Return all factors of a number."""
        return set(
            reduce(
                list.__add__,
                ([i, n // i] for i in range(1, int(n**0.5) + 1) if n % i == 0),
            )
        )

    def _find_supported_aspect_ratios(self):
        """
        This function computes all the allowed aspect ratios for a fixed
        number of input chunks.

        For example, with `num_tiles=5`, it will return:
        {
            0.2: [(1, 5)],
            5.0: [(5, 1)],
            0.25: [(1, 4)],
            1.0: [(2, 2), (1, 1)],
            4.0: [(4, 1)],
            0.3333333333333333: [(1, 3)],
            3.0: [(3, 1)],
            0.5: [(1, 2)],
            2.0: [(2, 1)]
        }
        """
        asp_dict = {}
        for chunk_size in range(self.max_num_tiles, 0, -1):
            _factors = sorted(VariableSizeImageTransform._factors(chunk_size))
            _asp_ratios = [(x, chunk_size // x) for x in _factors]
            for ratio in _asp_ratios:
                k = ratio[0] / ratio[1]
                if k not in asp_dict:
                    asp_dict[k] = [ratio]
                else:
                    asp_dict[k].append(ratio)
        return asp_dict

    def _find_closest_aspect_ratio(self, img_width: int, img_height: int) -> Tuple:
        """
        Given an image width, height and target number of chunks
        this function will find the closest supported aspect ratio.
        """
        tgt_ar = img_width / img_height
        asp_dict = self._find_supported_aspect_ratios()
        cl_d, cl_p = 1e23, None
        if tgt_ar >= 1:
            cl_p = min(
                [k for k in asp_dict.keys() if k <= tgt_ar],
                key=lambda x: abs(x - tgt_ar),
            )
            v = asp_dict[cl_p]
            # select width
            widths = [(idx, self.size * vv[0]) for idx, vv in enumerate(v)]
            tgt_idx = max(widths, key=lambda x: x[1])[0]
        else:
            cl_p = min(
                [k for k in asp_dict.keys() if k > tgt_ar],
                key=lambda x: abs(1 / x - 1 / tgt_ar),
            )
            v = asp_dict[cl_p]
            # select height
            heights = [(idx, self.size * vv[1]) for idx, vv in enumerate(v)]
            tgt_idx = max(heights, key=lambda x: x[1])[0]
        out = v[tgt_idx]
        return out

    def _resize(
        self, image: Image.Image, target_width: int, target_height: int
    ) -> Image.Image:
        # Resize longer edge to given size.
        w, h = image.size
        scale = w / h

        if scale > 1.0:
            # width > height
            new_w = target_width
            new_h = math.floor(new_w / scale)
        else:
            # height >= width
            new_h = target_height
            new_w = math.floor(new_h * scale)

        image = F.resize(image, (new_h, new_w))
        return image

    def _pad(self, image: Image.Image, new_width: int, new_height: int) -> Image.Image:
        mean_per_channel = tuple(
            np.clip(np.array(image).mean(axis=(0, 1)), 0, 255).astype(np.uint8)
        )
        new_im = Image.new(mode="RGB", size=(new_width, new_height), color=(0, 0, 0))  # type: ignore
        new_im.paste(image)
        return new_im

    def _split(self, image: torch.Tensor, ncw: int, nch: int) -> torch.Tensor:
        # Split image into number of required tiles (width x height)
        num_channels, height, width = image.size()
        image = image.view(num_channels, nch, height // nch, ncw, width // ncw)
        # Permute dimensions to reorder the axes
        image = image.permute(1, 3, 0, 2, 4).contiguous()
        # Reshape into the desired output shape (batch_size * 4, num_channels, width/2, height/2)
        image = image.view(ncw * nch, num_channels, height // nch, width // ncw)
        return image

    def _get_image_height_width(
        self, image_width: int, image_height: int, target_width: int, target_height: int
    ) -> Tuple[int, int]:
        """
        Given image width, height and target width, height for the canvas, return the dimensions of how the image would be resized
        with aspect ratio preservation.
        """
        scale = image_width / image_height

        if scale > 1.0:
            # Width is larger than height

            # Rescaling factor is the minimum of the two scaling factors. Else one side would be outside of the canvas.
            rescaling_factor = min(
                target_width / image_width, target_height / image_height
            )

            # Set new width to target width and height to the rescaled height.
            new_w = rescaling_factor * image_width
            new_h = math.floor(new_w / scale)

        else:
            # Height is larger than width

            # Rescaling factor is the minimum of the two scaling factors. Else one side would be outside of the canvas.
            rescaling_factor = min(
                target_width / image_width, target_height / image_height
            )

            # Set new height to target height and width to the rescaled width.
            new_h = rescaling_factor * image_height
            new_w = math.floor(new_h * scale)

        return new_w, new_h

    def _fit_image_to_canvas(
        self, img_width: int, img_height: int, area_limit: bool
    ) -> Any:
        """
        Given an image width, height and target number of chunks this function will see if the image
        can be fit into any of the canvases that can be build from arranging the tiles in a grid.
        If the image can be fit onto several canvases, it will return the canvas where the shorter edge
        of the image will be largest.

        If area_limit is set to True, the tie-breaking prefers the canvas where area is less than 2x the original area.
        """
        # Initialize the optimal canvas to None. If no canvas is found where image fits, function returns None.
        optimal_canvas = None
        optimal_image_width_height = None

        scale = img_width / img_height

        # Gather all potential supported image resolutions and iterate through them to find best match
        potential_arrangements = [
            item
            for sublist in self._find_supported_aspect_ratios().values()
            for item in sublist
        ]
        for n_w, n_h in potential_arrangements:
            # Compute the canvas size
            canvas_width, canvas_height = n_w * self.size, n_h * self.size

            # Check if image can fit into the canvas without downsampling
            if canvas_width >= img_width and canvas_height >= img_height:
                # If we did not find a good canvas yet, we will use the current one
                if optimal_canvas is None:
                    # Set optimal canvas and determine the actual image height and width in the canvas with aspect ratio preserving resampling
                    optimal_canvas = (n_w, n_h)
                    optimal_image_width_height = self._get_image_height_width(
                        image_width=img_width,
                        image_height=img_height,
                        target_width=n_w * self.size,
                        target_height=n_h * self.size,
                    )
                else:
                    # If we already found an optimal canvas before, we will check if the shorter edge of the image will be larger than the current optimal canvas.
                    # This means we can potentially upsample the image resolution which is beneficial to performance.
                    image_width_height = self._get_image_height_width(
                        image_width=img_width,
                        image_height=img_height,
                        target_width=n_w * self.size,
                        target_height=n_h * self.size,
                    )
                    if area_limit:
                        # Prioritize aspect ratio, and choose best within area limit when tied.
                        curr_scale = image_width_height[0] / image_width_height[1]
                        optim_scale = (
                            optimal_image_width_height[0]
                            / optimal_image_width_height[1]
                        )
                        if abs(scale - curr_scale) < abs(scale - optim_scale):
                            # 1. optimize aspect ratio
                            optimal_canvas = (n_w, n_h)
                            optimal_image_width_height = image_width_height
                        elif abs(scale - curr_scale) == abs(scale - optim_scale):
                            # 2. optimize area
                            if (
                                image_width_height[0] * image_width_height[1]
                                < 2 * img_width * img_height
                            ):
                                # 2.1 area is less than 2x the original area
                                optimal_canvas = (n_w, n_h)
                                optimal_image_width_height = image_width_height
                    else:
                        # NOTE: L3V dynamid tiling. Priortize biggest canvas.
                        if (
                            scale < 1.0
                            and (image_width_height[0] >= optimal_image_width_height[0])
                        ) or (
                            scale >= 1.0
                            and (image_width_height[1] >= optimal_image_width_height[1])
                        ):
                            optimal_canvas = (n_w, n_h)
                            optimal_image_width_height = image_width_height
        return optimal_canvas

    def __call__(self, image: Image.Image) -> Tuple[Any, Any]:
        assert isinstance(image, Image.Image), type(image)
        if self.use_thumbnail != "no":
            thumbnail = self.thumbnail_transform(image)[0]

        w, h = image.size
        # Check if the image can be fit to the canvas without downsampling
        ar = self._fit_image_to_canvas(
            img_width=w, img_height=h, area_limit=self.area_limit
        )
        if ar is None:
            # If we did not find a canvas, we have to find the closest aspect ratio and downsample the image
            ar = self._find_closest_aspect_ratio(img_width=w, img_height=h)

        image = F.resize(
            image,
            (ar[1] * self.size, ar[0] * self.size),  # (h, w)
            interpolation=InterpolationMode.BICUBIC,
        )
        image = self._pad(image, ar[0] * self.size, ar[1] * self.size)
        image = self.to_tensor(image)
        image = self.normalize(image)
        image = self._split(image, ar[0], ar[1])  # type: ignore
        if self.use_thumbnail == "before":
            image = torch.cat((thumbnail, image), dim=0)
        elif self.use_thumbnail == "after":
            image = torch.cat((image, thumbnail), dim=0)
        elif self.use_thumbnail == "both":
            image = torch.cat((thumbnail, image, thumbnail), dim=0)

        return image, ar



"""KID (Kernel Inception Distance) module.

KID measures the distance between distributions of generated and reference images
through Maximum Mean Discrepancy (MMD) in the Inception feature space. Unlike FID,
KID has an unbiased estimator and works correctly on small sample sizes (from ~50
images). Lower KID = better generation quality. Typical range: 0.0-0.1.

This is a dataset-level metric that compares two distributions of images/videos.
"""

import logging
from pathlib import Path
from typing import Optional, List, Tuple

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.base_modules import BatchMetricModule

logger = logging.getLogger(__name__)


class KIDModule(BatchMetricModule):
    name = "kid"
    description = "Kernel Inception Distance for image generation evaluation (batch metric)"
    default_config = {
        "feature_layer": "2048",  # Inception feature layer
        "subset_size": 100,  # Subset size for KID estimation
        "num_subsets": 100,  # Number of subsets for averaging
        "degree": 3,  # Polynomial kernel degree
        "gamma": None,  # Kernel gamma (None = auto: 1/feature_dim)
        "coef0": 1.0,  # Polynomial kernel coefficient
        "device": "auto",
        "batch_size": 32,
        "resize": 299,  # Input size for InceptionV3
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subset_size = self.config.get("subset_size", 100)
        self.num_subsets = self.config.get("num_subsets", 100)
        self.degree = self.config.get("degree", 3)
        self.gamma = self.config.get("gamma", None)
        self.coef0 = self.config.get("coef0", 1.0)
        self.device_config = self.config.get("device", "auto")
        self.batch_size = self.config.get("batch_size", 32)
        self.resize = self.config.get("resize", 299)
        self.device = None
        self._ml_available = False
        self._backend = None  # "cleanfid", "torch_fidelity", or "native"
        self._inception_model = None
        self._transform = None

    def setup(self) -> None:
        # Tier 1: clean-fid
        try:
            import cleanfid  # noqa: F401
            self._backend = "cleanfid"
            self._ml_available = True
            logger.info("KID module initialized with clean-fid backend")
            return
        except ImportError:
            pass

        # Tier 2: torch-fidelity
        try:
            import torch_fidelity  # noqa: F401
            self._backend = "torch_fidelity"
            self._ml_available = True
            logger.info("KID module initialized with torch-fidelity backend")
            return
        except ImportError:
            pass

        # Tier 3: Native InceptionV3 + polynomial kernel MMD
        try:
            import torch
            from torchvision import models, transforms
            from torchvision.models import Inception_V3_Weights

            if self.device_config == "auto":
                self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            else:
                self.device = torch.device(self.device_config)

            # Load InceptionV3 for feature extraction
            model = models.inception_v3(
                weights=Inception_V3_Weights.IMAGENET1K_V1,
                transform_input=False,
            )
            # Remove final classification layer to get 2048-d features
            model.fc = torch.nn.Identity()
            model = model.to(self.device)
            model.eval()
            self._inception_model = model

            self._transform = transforms.Compose([
                transforms.ToPILImage(),
                transforms.Resize((self.resize, self.resize)),
                transforms.ToTensor(),
                transforms.Normalize(
                    mean=[0.485, 0.456, 0.406],
                    std=[0.229, 0.224, 0.225],
                ),
            ])

            self._backend = "native"
            self._ml_available = True
            logger.info(f"KID module initialized with native InceptionV3 on {self.device}")

        except ImportError as e:
            logger.warning(f"Missing dependencies for KID (torch/torchvision required): {e}")
        except Exception as e:
            logger.warning(f"Failed to setup KID: {e}")

    def extract_features(self, sample: Sample) -> Optional[np.ndarray]:
        """Extract InceptionV3 features from a sample image or video frame.

        Args:
            sample: Sample to extract features from

        Returns:
            Feature vector (numpy array of shape (2048,)), or None if failed
        """
        if not self._ml_available or self._backend != "native":
            # For cleanfid/torch-fidelity backends, features are computed in bulk
            # at on_dispose time. Cache the path instead.
            return str(sample.path) if self._ml_available else None

        try:
            import torch

            # Load representative frame
            frame = self._load_frame(sample)
            if frame is None:
                return None

            # Convert BGR to RGB
            frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

            # Transform and extract features
            tensor = self._transform(frame_rgb).unsqueeze(0).to(self.device)

            with torch.no_grad():
                features = self._inception_model(tensor)
                # inception_v3 may return InceptionOutputs during eval
                if hasattr(features, "logits"):
                    features = features.logits
                features = features.cpu().numpy().flatten()

            return features

        except Exception as e:
            logger.debug(f"Failed to extract Inception features from {sample.path}: {e}")
            return None

    def _load_frame(self, sample: Sample) -> Optional[np.ndarray]:
        """Load a representative frame from an image or video.

        For images: loads the image directly.
        For videos: extracts the middle frame.

        Returns:
            BGR frame as numpy array, or None if loading failed
        """
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total_frames <= 0:
                    cap.release()
                    return None
                # Sample middle frame
                cap.set(cv2.CAP_PROP_POS_FRAMES, total_frames // 2)
                ret, frame = cap.read()
                cap.release()
                return frame if ret else None
            else:
                frame = cv2.imread(str(sample.path))
                return frame
        except Exception as e:
            logger.debug(f"Failed to load frame: {e}")
            return None

    def _polynomial_kernel(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """Compute polynomial kernel K(x, y) = (gamma * x.y + coef0)^degree.

        Args:
            x: Features array (n, d)
            y: Features array (m, d)

        Returns:
            Kernel matrix (n, m)
        """
        d = x.shape[1]
        gamma = self.gamma if self.gamma is not None else 1.0 / d
        return (gamma * (x @ y.T) + self.coef0) ** self.degree

    def _compute_kid_mmd(
        self, features: np.ndarray, ref_features: np.ndarray
    ) -> Tuple[float, float]:
        """Compute KID via polynomial kernel MMD with subsampling.

        Args:
            features: Generated features (N, d)
            ref_features: Reference features (M, d)

        Returns:
            Tuple of (mean KID, std KID) across subsets
        """
        n = min(len(features), len(ref_features), self.subset_size)
        if n < 2:
            return float("inf"), 0.0

        kid_values = []
        for _ in range(self.num_subsets):
            # Random subsets
            idx_x = np.random.choice(len(features), size=n, replace=False)
            idx_y = np.random.choice(len(ref_features), size=n, replace=False)

            x = features[idx_x]
            y = ref_features[idx_y]

            # MMD^2 = E[k(x,x')] + E[k(y,y')] - 2*E[k(x,y)]
            kxx = self._polynomial_kernel(x, x)
            kyy = self._polynomial_kernel(y, y)
            kxy = self._polynomial_kernel(x, y)

            # Unbiased estimator: exclude diagonal
            np.fill_diagonal(kxx, 0)
            np.fill_diagonal(kyy, 0)

            m = n  # subset size
            mmd2 = (
                kxx.sum() / (m * (m - 1))
                + kyy.sum() / (m * (m - 1))
                - 2.0 * kxy.sum() / (m * m)
            )
            kid_values.append(float(mmd2))

        return float(np.mean(kid_values)), float(np.std(kid_values))

    def compute_distribution_metric(
        self, features: List, reference_features: Optional[List] = None
    ) -> float:
        """Compute KID between feature distributions.

        Args:
            features: List of feature vectors (or paths for cleanfid/torch-fidelity)
            reference_features: Optional list of reference features

        Returns:
            KID score (lower is better)
        """
        if self._backend == "cleanfid":
            return self._compute_cleanfid(features, reference_features)
        elif self._backend == "torch_fidelity":
            return self._compute_torch_fidelity(features, reference_features)
        else:
            return self._compute_native(features, reference_features)

    def _compute_cleanfid(self, features: List, reference_features: Optional[List]) -> float:
        """Compute KID using clean-fid library."""
        try:
            from cleanfid import fid as cleanfid_module

            # clean-fid expects directories of images
            # features here are paths (strings)
            if reference_features and len(reference_features) > 0:
                # Create temp dirs with symlinks
                import tempfile
                import os

                with tempfile.TemporaryDirectory() as gen_dir, \
                     tempfile.TemporaryDirectory() as ref_dir:

                    for i, p in enumerate(features):
                        src = Path(p)
                        if src.exists():
                            dst = Path(gen_dir) / f"{i}{src.suffix}"
                            try:
                                os.symlink(src, dst)
                            except OSError:
                                import shutil
                                shutil.copy2(src, dst)

                    for i, p in enumerate(reference_features):
                        src = Path(p)
                        if src.exists():
                            dst = Path(ref_dir) / f"{i}{src.suffix}"
                            try:
                                os.symlink(src, dst)
                            except OSError:
                                import shutil
                                shutil.copy2(src, dst)

                    score = cleanfid_module.compute_kid(gen_dir, ref_dir)
                    return float(score)

            logger.warning("KID (clean-fid): no reference features, cannot compute")
            return float("inf")

        except Exception as e:
            logger.error(f"Failed to compute KID via clean-fid: {e}")
            return float("inf")

    def _compute_torch_fidelity(
        self, features: List, reference_features: Optional[List]
    ) -> float:
        """Compute KID using torch-fidelity library."""
        try:
            import torch_fidelity

            if reference_features and len(reference_features) > 0:
                import tempfile
                import os

                with tempfile.TemporaryDirectory() as gen_dir, \
                     tempfile.TemporaryDirectory() as ref_dir:

                    for i, p in enumerate(features):
                        src = Path(p)
                        if src.exists():
                            dst = Path(gen_dir) / f"{i}{src.suffix}"
                            try:
                                os.symlink(src, dst)
                            except OSError:
                                import shutil
                                shutil.copy2(src, dst)

                    for i, p in enumerate(reference_features):
                        src = Path(p)
                        if src.exists():
                            dst = Path(ref_dir) / f"{i}{src.suffix}"
                            try:
                                os.symlink(src, dst)
                            except OSError:
                                import shutil
                                shutil.copy2(src, dst)

                    metrics = torch_fidelity.calculate_metrics(
                        input1=gen_dir,
                        input2=ref_dir,
                        kid=True,
                        kid_subset_size=self.subset_size,
                        kid_subsets=self.num_subsets,
                    )
                    return float(metrics.get("kernel_inception_distance_mean", float("inf")))

            logger.warning("KID (torch-fidelity): no reference features, cannot compute")
            return float("inf")

        except Exception as e:
            logger.error(f"Failed to compute KID via torch-fidelity: {e}")
            return float("inf")

    def _compute_native(
        self, features: List[np.ndarray], reference_features: Optional[List[np.ndarray]]
    ) -> float:
        """Compute KID using native InceptionV3 + polynomial kernel MMD."""
        try:
            features_array = np.stack(features, axis=0)

            if reference_features is not None and len(reference_features) > 0:
                ref_array = np.stack(reference_features, axis=0)
            else:
                # Split features in half for self-comparison
                mid = len(features_array) // 2
                if mid < 2:
                    return float("inf")
                ref_array = features_array[:mid]
                features_array = features_array[mid:]

            kid_mean, kid_std = self._compute_kid_mmd(features_array, ref_array)

            logger.info(f"KID computed: {kid_mean:.6f} +/- {kid_std:.6f}")

            # Store std via pipeline if available
            if hasattr(self, "pipeline") and self.pipeline:
                if hasattr(self.pipeline, "add_dataset_metric"):
                    self.pipeline.add_dataset_metric("kid_std", kid_std)

            return kid_mean

        except Exception as e:
            logger.error(f"Failed to compute KID (native): {e}")
            return float("inf")

    def process(self, sample: Sample) -> Sample:
        """Extract and cache features from sample.

        Does not modify the sample directly. Features are accumulated for
        batch computation in on_dispose().
        """
        if not self._ml_available:
            return sample

        features = self.extract_features(sample)
        if features is not None:
            self._feature_cache.append(features)

        # Check if sample has reference for paired comparison
        reference_path = getattr(sample, "reference_path", None)
        if reference_path is not None and isinstance(reference_path, (str, Path)):
            try:
                ref_path = Path(reference_path) if isinstance(reference_path, str) else reference_path
                if ref_path.exists():
                    ref_sample = Sample(
                        path=ref_path,
                        is_video=sample.is_video,
                    )
                    ref_features = self.extract_features(ref_sample)
                    if ref_features is not None:
                        self._reference_cache.append(ref_features)
            except Exception as e:
                logger.debug(f"Failed to extract reference features: {e}")

        return sample

    def on_dispose(self) -> None:
        """Compute KID after all samples processed."""
        if len(self._feature_cache) < 2:
            logger.info(
                f"KID: Not enough samples ({len(self._feature_cache)}) for metric computation"
            )
            self._feature_cache = []
            self._reference_cache = []
            return

        try:
            kid_score = self.compute_distribution_metric(
                self._feature_cache,
                self._reference_cache if self._reference_cache else None,
            )

            logger.info(
                f"KID computed: {kid_score:.6f} "
                f"(generated: {len(self._feature_cache)}, "
                f"reference: {len(self._reference_cache)})"
            )

            # Store in pipeline stats
            if hasattr(self, "pipeline") and self.pipeline:
                if hasattr(self.pipeline, "add_dataset_metric"):
                    self.pipeline.add_dataset_metric("kid", kid_score)

        except Exception as e:
            logger.error(f"Failed to compute KID: {e}")

        finally:
            self._feature_cache = []
            self._reference_cache = []



"""Tests for faver module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_faver_basics():
    from ayase.modules.faver import FAVERModule
    _test_module_basics(FAVERModule, "faver")

def test_faver_video(video_sample):
    from ayase.modules.faver import FAVERModule
    video_sample.quality_metrics = QualityMetrics()
    m = FAVERModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for background_consistency module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_background_consistency_basics():
    from ayase.modules.background_consistency import BackgroundConsistencyModule
    _test_module_basics(BackgroundConsistencyModule, "background_consistency")

def test_background_consistency_video(video_sample):
    from ayase.modules.background_consistency import BackgroundConsistencyModule
    video_sample.quality_metrics = QualityMetrics()
    m = BackgroundConsistencyModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""LMM (Large Multimodal Model) Descriptive Quality Assessment module.

Uses large multimodal models (e.g., LLaVA-NeXT, GPT-4V) to generate
natural language quality explanations. Provides interpretable quality
assessment with specific issue identification and reasoning.
"""

import logging
from pathlib import Path
from typing import Optional, List

import cv2
import numpy as np

from ayase.models import Sample, ValidationIssue, ValidationSeverity, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class LLMDescriptiveQAModule(PipelineModule):
    name = "llm_descriptive_qa"
    description = "LMM-based interpretable quality assessment with explanations"
    default_config = {
        "model_name": "llava-hf/llava-v1.6-mistral-7b-hf",  # LLaVA-NeXT
        "use_openai": False,  # Use OpenAI GPT-4V instead
        "openai_api_key": None,  # OpenAI API key if using GPT-4V
        "num_frames": 4,  # Frames to analyze
        "device": "auto",
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.model_name = self.config.get("model_name", "llava-hf/llava-v1.6-mistral-7b-hf")
        self.use_openai = self.config.get("use_openai", False)
        self.openai_api_key = self.config.get("openai_api_key", None)
        self.num_frames = self.config.get("num_frames", 4)
        self.device_config = self.config.get("device", "auto")
        self.device = None
        self._ml_available = False
        self._model = None
        self._processor = None

    def setup(self) -> None:
        try:
            if self.use_openai:
                # OpenAI API setup
                if self.openai_api_key:
                    self._ml_available = True
                    logger.info("LLM Descriptive QA initialized with OpenAI API")
                else:
                    logger.warning("OpenAI API key not provided")
            else:
                # Local LLaVA model setup
                from transformers import AutoProcessor, LlavaForConditionalGeneration
                import torch

                # Set device
                if self.device_config == "auto":
                    self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                else:
                    self.device = torch.device(self.device_config)

                logger.info(f"Loading LLaVA model: {self.model_name}")

                self._processor = AutoProcessor.from_pretrained(self.model_name)
                self._model = LlavaForConditionalGeneration.from_pretrained(
                    self.model_name,
                    torch_dtype=torch.float16 if torch.cuda.is_available() else torch.float32,
                    low_cpu_mem_usage=True,
                ).to(self.device)

                self._ml_available = True
                logger.info(f"LLM Descriptive QA initialized with LLaVA on {self.device}")

        except ImportError:
            logger.warning(
                "transformers or torch not installed. "
                "Install with: pip install transformers torch"
            )
        except Exception as e:
            logger.warning(f"Failed to setup LLM Descriptive QA: {e}")

    def _load_key_frames(self, video_path: Path) -> Optional[List[np.ndarray]]:
        """Load key frames from video for analysis."""
        try:
            cap = cv2.VideoCapture(str(video_path))
            total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

            if total_frames == 0:
                cap.release()
                return None

            # Sample frames uniformly
            num_frames = min(self.num_frames, total_frames)
            frame_indices = np.linspace(0, total_frames - 1, num_frames, dtype=int)
            frames = []

            for idx in frame_indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    # Convert BGR to RGB
                    frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    frames.append(frame_rgb)

            cap.release()
            return frames if frames else None

        except Exception as e:
            logger.debug(f"Failed to load frames: {e}")
            return None

    def _analyze_with_llava(self, frames: List[np.ndarray]) -> Optional[dict]:
        """Analyze frames using local LLaVA model."""
        try:
            import torch
            from PIL import Image

            # Use middle frame for analysis
            frame = frames[len(frames) // 2]
            image = Image.fromarray(frame)

            # Quality assessment prompt
            prompt = (
                "USER: <image>\nAnalyze the quality of this video frame. "
                "Identify any issues with: sharpness, motion blur, lighting, "
                "artifacts, color, consistency. "
                "Rate the overall quality from 0-100 and explain your assessment.\nASSISTANT:"
            )

            # Process input
            inputs = self._processor(text=prompt, images=image, return_tensors="pt").to(self.device)

            # Generate response
            with torch.no_grad():
                output = self._model.generate(
                    **inputs,
                    max_new_tokens=200,
                    do_sample=False,
                )

            # Decode response
            response = self._processor.decode(output[0], skip_special_tokens=True)

            # Extract quality score and explanation
            # Simple parsing (production would need better extraction)
            explanation = response.split("ASSISTANT:")[-1].strip()

            # Try to extract numerical score
            import re
            # Prefer explicit N/100 pattern
            score_match = re.search(r'(\d+)\s*/\s*100', explanation)
            if score_match:
                quality_score = float(score_match.group(1))
            else:
                # Fall back to last number found
                all_numbers = re.findall(r'(\d+)', explanation)
                quality_score = float(all_numbers[-1]) if all_numbers else 50.0

            return {
                "quality_score": quality_score,
                "explanation": explanation,
                "issues": self._extract_issues(explanation),
            }

        except Exception as e:
            logger.warning(f"LLaVA analysis failed: {e}")
            return None

    def _analyze_with_openai(self, frames: List[np.ndarray]) -> Optional[dict]:
        """Analyze frames using OpenAI GPT-4V API."""
        try:
            from openai import OpenAI
            import base64
            from io import BytesIO
            from PIL import Image

            client = OpenAI(api_key=self.openai_api_key)

            # Use middle frame
            frame = frames[len(frames) // 2]
            image = Image.fromarray(frame)

            # Convert to base64
            buffered = BytesIO()
            image.save(buffered, format="PNG")
            img_base64 = base64.b64encode(buffered.getvalue()).decode()

            # API call
            response = client.chat.completions.create(
                model="gpt-4o",
                messages=[
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "text",
                                "text": "Analyze the quality of this video frame. "
                                "Identify any issues with: sharpness, motion, lighting, "
                                "artifacts, color. Rate quality 0-100 and explain.",
                            },
                            {
                                "type": "image_url",
                                "image_url": {"url": f"data:image/png;base64,{img_base64}"},
                            },
                        ],
                    }
                ],
                max_tokens=300,
            )

            explanation = response.choices[0].message.content

            # Extract quality score
            import re
            score_match = re.search(r'(\d+)\s*(?:/\s*100)?', explanation)
            quality_score = float(score_match.group(1)) if score_match else 50.0

            return {
                "quality_score": quality_score,
                "explanation": explanation,
                "issues": self._extract_issues(explanation),
            }

        except Exception as e:
            logger.warning(f"OpenAI analysis failed: {e}")
            return None

    def _extract_issues(self, explanation: str) -> List[str]:
        """Extract quality issues from explanation text."""
        issues = []
        issue_keywords = ["blur", "artifact", "noise", "dark", "bright", "saturated", "compressed"]

        explanation_lower = explanation.lower()
        for keyword in issue_keywords:
            if keyword in explanation_lower:
                issues.append(keyword)

        return issues

    def process(self, sample: Sample) -> Sample:
        """Process sample with LMM descriptive quality assessment."""
        if not self._ml_available:
            return sample

        try:
            # Load frames
            if sample.is_video:
                frames = self._load_key_frames(sample.path)
            else:
                # Load single image
                img = cv2.imread(str(sample.path))
                if img is not None:
                    frames = [cv2.cvtColor(img, cv2.COLOR_BGR2RGB)]
                else:
                    frames = None

            if frames is None:
                return sample

            # Analyze with LMM
            if self.use_openai:
                result = self._analyze_with_openai(frames)
            else:
                result = self._analyze_with_llava(frames)

            if result is None:
                return sample

            # Store confidence score (LMM's quality rating)
            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.llm_qa_score = result["quality_score"] / 100.0

            # Add explanation as validation issue
            sample.validation_issues.append(
                ValidationIssue(
                    severity=ValidationSeverity.INFO,
                    message="LMM Quality Assessment",
                    details={
                        "explanation": result["explanation"],
                        "detected_issues": result["issues"],
                        "quality_score": result["quality_score"],
                    },
                    recommendation=result["explanation"],
                )
            )

            logger.debug(
                f"LMM analysis for {sample.path.name}: {result['quality_score']:.1f}/100"
            )

        except Exception as e:
            logger.warning(f"LLM Descriptive QA processing failed for {sample.path}: {e}")

        return sample



"""Face-IQA module using TOPIQ face-specific variant."""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class FaceIQAModule(PipelineModule):
    name = "face_iqa"
    description = "Face-specific IQA via TOPIQ-face (GFIQA-trained, higher=better)"
    default_config = {"subsample": 8}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._model = None
        self._face_cascade = None

    def setup(self) -> None:
        try:
            import pyiqa
            import torch

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self._model = pyiqa.create_metric("topiq_nr-face", device=device)
            try:
                self._device = next(self._model.parameters()).device
            except StopIteration:
                self._device = torch.device("cpu")
            self._ml_available = True
            logger.info("Face-IQA (topiq_nr-face) model loaded on %s", device)
        except (ImportError, Exception) as e:
            logger.warning("Face-IQA unavailable: %s", e)

        try:
            import cv2

            cascade_path = cv2.data.haarcascades + "haarcascade_frontalface_default.xml"
            self._face_cascade = cv2.CascadeClassifier(cascade_path)
        except Exception:
            pass

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        if not self._ml_available:
            return sample
        try:
            import cv2
            import torch

            frames = self._load_frames(sample)
            if not frames:
                return sample

            face_scores = []
            device = self._device

            for frame in frames:
                faces = self._detect_faces(frame)
                if not faces:
                    continue

                for (x, y, w, h) in faces:
                    # Pad face region by 20%
                    pad = int(max(w, h) * 0.2)
                    fy = max(0, y - pad)
                    fx = max(0, x - pad)
                    fh = min(frame.shape[0], y + h + pad) - fy
                    fw = min(frame.shape[1], x + w + pad) - fx
                    face_crop = frame[fy : fy + fh, fx : fx + fw]

                    if face_crop.size == 0:
                        continue

                    rgb = cv2.cvtColor(face_crop, cv2.COLOR_BGR2RGB)
                    tensor = (
                        torch.from_numpy(rgb)
                        .permute(2, 0, 1)
                        .unsqueeze(0)
                        .float()
                        / 255.0
                    )
                    tensor = tensor.to(device)
                    with torch.no_grad():
                        score = self._model(tensor).item()
                    face_scores.append(score)

            if face_scores:
                sample.quality_metrics.face_iqa_score = float(np.mean(face_scores))
        except Exception as e:
            logger.warning("Face-IQA processing failed: %s", e)
        return sample

    def _detect_faces(self, frame) -> list:
        if self._face_cascade is None:
            return []
        import cv2

        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        faces = self._face_cascade.detectMultiScale(gray, 1.1, 4, minSize=(30, 30))
        return list(faces) if len(faces) > 0 else []

    def _load_frames(self, sample: Sample) -> list:
        import cv2

        subsample = self.config.get("subsample", 8)
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = max(int(cap.get(cv2.CAP_PROP_FRAME_COUNT)), 0)
            indices = list(range(0, total, max(1, total // subsample)))[:subsample]
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is not None:
                frames.append(frame)
        return frames



"""MC360IQA — Multi-Channel Blind 360° IQA. mc360iqa_score — higher = better"""
import logging, cv2, numpy as np
from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule
logger = logging.getLogger(__name__)
class MC360IQAModule(PipelineModule):
    name = "mc360iqa"; description = "MC360IQA blind 360 IQA (2019)"; default_config = {"subsample": 8, "n_viewports": 6}
    def __init__(self, c=None):
        super().__init__(c); self.subsample = self.config.get("subsample", 8); self._backend = "heuristic"
    def setup(self):
        try: import mc360iqa; self._model = mc360iqa; self._backend = "native"; return
        except ImportError: pass
        self._backend = "heuristic"
    def process(self, sample):
        try:
            frames = self._ex(sample)
            if not frames: return sample
            # Multi-viewport quality sampling
            scores = []
            for f in frames:
                h, w = f.shape[:2]
                vp_size = min(h, w) // 3
                if vp_size < 32: vp_size = min(h, w)
                vp_scores = []
                for y_frac, x_frac in [(0.25,0.25),(0.25,0.75),(0.5,0.5),(0.75,0.25),(0.75,0.75),(0.5,0.1)]:
                    cy, cx = int(h*y_frac), int(w*x_frac)
                    y1, x1 = max(cy-vp_size//2,0), max(cx-vp_size//2,0)
                    vp = f[y1:y1+vp_size, x1:x1+vp_size]
                    if vp.size == 0: continue
                    gray = cv2.cvtColor(vp, cv2.COLOR_BGR2GRAY).astype(np.float64)
                    vp_scores.append(min(cv2.Laplacian(gray, cv2.CV_64F).var()/500, 1))
                if vp_scores: scores.append(np.mean(vp_scores))
            if sample.quality_metrics is None: sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.mc360iqa_score = float(np.clip(np.mean(scores), 0, 1)) if scores else None
        except Exception as e: logger.warning(f"MC360IQA failed: {e}")
        return sample
    def _ex(self, sample):
        frames=[]
        if sample.is_video:
            cap=cv2.VideoCapture(str(sample.path)); tot=int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if tot<=0: cap.release(); return []
            for i in np.linspace(0,tot-1,min(self.subsample,tot),dtype=int):
                cap.set(cv2.CAP_PROP_POS_FRAMES,i); r,f=cap.read()
                if r: frames.append(f)
            cap.release()
        else:
            img=cv2.imread(str(sample.path))
            if img is not None: frames.append(img)
        return frames



# Copyright (c) Meta Platforms, Inc. and affiliates.
# This software may be used and distributed according to the terms of the Llama 2 Community License Agreement.

import contextlib
import logging
import os
from dataclasses import dataclass
from pathlib import Path

import torch.distributed
import wandb
import xformers.profiler
from torch.profiler.profiler import profile
from xformers.profiler import MemSnapshotsProfiler, PyTorchProfiler

from ..core.distributed import get_is_master


@dataclass
class ProfilerArgs:
    run: bool = False
    trace_folder: str = "profiling"
    mem_warmup: int = 100
    mem_steps: int = 2
    profile_warmup: int = 102
    profile_steps: int = 2


logger = logging.getLogger()


def perfetto_to_html(json_file, html_file):
    import gzip
    import string

    import viztracer

    root = os.path.dirname(viztracer.__file__)
    sub = {}
    json_file = gzip.open(json_file) if ".gz" in str(json_file) else open(json_file)
    with open(
        os.path.join(root, "html/trace_viewer_embedder.html"), encoding="utf-8"
    ) as f:
        tmpl = f.read()
    with open(os.path.join(root, "html/trace_viewer_full.html"), encoding="utf-8") as f:
        sub["trace_viewer_full"] = f.read()
    with json_file as j:
        content = j.read()
        if isinstance(content, bytes):
            content = content.decode("utf-8")
        sub["json_data"] = content.replace("</script>", "<\\/script>")  # type: ignore
    with open(html_file, "w+", encoding="utf-8") as output_file:
        output_file.write(string.Template(tmpl).substitute(sub))


class PyTorchProfilerWandb(PyTorchProfiler):
    def __init__(self, main_profiler) -> None:
        self.main_profiler = main_profiler
        self.num_steps = 0
        self.pytorch_profiler = torch.profiler.profile(
            on_trace_ready=self._on_trace,
            profile_memory=True,
            record_shapes=True,
            # With stack gives huge profile traces
            # and bugs out because of some non ascii
            # character somewhere in pytorch
            with_stack=False,
            with_flops=True,
            activities=self.ACTIVITIES,
        )

    def _analyze_trace(self, prof: profile):
        logger.info("Begin analyze trace")
        super()._analyze_trace(prof)
        logger.info("End analyze trace")

    def _on_trace(self, prof: torch.profiler.profiler.profile) -> None:
        super()._on_trace(prof)
        if get_is_master() and wandb.run is not None:
            filename = list(
                Path(self.main_profiler.output_dir).glob(
                    "profile_CPU_CUDA*/*.pt.trace.json*"
                )
            )[0]
            html_path = str(filename).replace(".json", ".html")
            perfetto_to_html(filename, html_path)
            wandb.log({"profile_trace": wandb.Html(html_path)})


class MemSnapshotsProfilerWandb(MemSnapshotsProfiler):
    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        if get_is_master() and wandb.run is not None:
            filename = list(
                Path(self.main_profiler.output_dir).glob("memory_trace_plot/*.html")
            )[0]
            wandb.log({"memory_trace": wandb.Html(open(filename), inject=False)})


@contextlib.contextmanager
def maybe_run_profiler(dump_dir, module, config: ProfilerArgs):
    # get user defined profiler settings

    if config.run:
        trace_dir = os.path.join(dump_dir, config.trace_folder)

        logger.info(f"Profiling active.  Traces will be saved at {trace_dir}")

        if get_is_master() and not os.path.exists(trace_dir):
            os.makedirs(trace_dir)
        if torch.distributed.is_initialized():
            torch.distributed.barrier()

        with xformers.profiler.profile(
            output_dir=trace_dir,
            module=module,
            schedule=[
                (
                    MemSnapshotsProfilerWandb,
                    config.mem_warmup,
                    config.mem_warmup + config.mem_steps,
                ),
                (
                    PyTorchProfilerWandb,
                    config.profile_warmup,
                    config.profile_warmup + config.profile_steps,
                ),
            ],
        ) as profiler:
            yield profiler

    else:
        torch_profiler = contextlib.nullcontext()
        yield None



import torch
from fvcore.nn import FlopCountAnalysis, flop_count_table
from torch.nn import MultiheadAttention

from models.beit.st_beit import BeitConfig, BeitModel
from models.temporal_model import (STAdapter, TemporalAttention,
                                   WindowTemporalAttention)


def mem_stat():
    mem = torch.cuda.max_memory_allocated() / 1024 / 1024
    print(f"max memory allocated: {mem}MB")


def build_backbone(tm_block="timesformer"):
    """TODO: Docstring for build_backbone.
    Returns: TODO

    """
    if tm_block == "timesformer":
        other_cfg = dict(
            num_frames=12, temporal_model_block="timesformer", temporal_model_config={}
        )
    elif tm_block == "st_adapter":
        other_cfg = dict(
            num_frames=12, temporal_model_block="st_adapter", temporal_model_config={}
        )
    elif tm_block == "xclip":
        other_cfg = dict(
            num_frames=12, temporal_model_block="xclip", temporal_model_config={}
        )
    elif tm_block == "none":
        other_cfg = dict(num_frames=12, temporal_model_block="none", temporal_model_config={})
    elif tm_block == "wa_2x2":
        other_cfg = dict(
            num_frames=12,
            temporal_model_block="window_attention",
            temporal_model_config=dict(window_size=(2, 2)),
        )
    elif tm_block == "wa_7x7":
        other_cfg = dict(
            num_frames=12,
            temporal_model_block="window_attention",
            temporal_model_config=dict(window_size=(7, 7)),
        )
    else:
        raise ValueError("not exist")

    model_card = "microsoft/beit-base-patch16-224-pt22k-ft22k"
    model_config = BeitConfig.from_pretrained(model_card, image_size=224, **other_cfg)
    model = BeitModel(model_config)
    return model


# model = TemporalAttention()
model = build_backbone("st_adapter")
model.gradient_checkpointing_enable()
model.cuda()
for i in range(3):
    x = torch.rand(32, 12, 3, 224, 224, requires_grad=True)
    x = x.cuda()
    x = x.requires_grad_()
    y = model(x)
    loss = y[0].mean()
    loss.backward()
    mem_stat()

# flops = FlopCountAnalysis(model, x)
# print(flop_count_table(flops))



import datetime
import logging
import time
from os.path import join

import pandas as pd
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import wandb
from torch.utils.data import ConcatDataset

from dataset.serialize import local_broadcast_process_authkey
from dataset import MetaLoader_rs, create_dataset, create_loader, create_sampler, create_stateful_sampler
from models import *
from tasks_clip.retrieval_utils import evaluation_wrapper
from tasks_clip.shared_utils import get_media_types, setup_model
from utils.basic_utils import MetricLogger, SmoothedValue, setup_seed
from utils.config_utils import setup_main
from utils.distributed import get_rank, is_main_process
from utils.logger import log_dict_to_wandb, setup_wandb

logger = logging.getLogger(__name__)


def train(
    model,
    train_loaders,
    optimizer,
    tokenizer,
    epoch,
    global_step,
    device,
    scheduler,
    scaler,
    config,
    data_type,
    skip_num=0
):
    model.train()

    metric_logger = MetricLogger(delimiter="  ")
    metric_logger.add_meter("lr", SmoothedValue(window=100, fmt="{value:.6f}"))
    metric_logger.add_meter("temperature", SmoothedValue(window=100, fmt="{value:.4f}"))
    loss_names = ["loss_" + k for k, v in config.criterion.loss_weight.items() if v != 0]

    media_types = get_media_types(train_loaders)

    for name in loss_names:
        for m in media_types:
            metric_logger.add_meter(
                f"{m}-{name}", SmoothedValue(window=100, fmt="{value:.4f}")
            )

    header = f"Train Epoch: [{epoch}]"
    log_freq = config.log_freq

    if config.distributed:
        for d in train_loaders:
            d.sampler.set_epoch(epoch)
    train_loader = MetaLoader_rs(name2loader=dict(list(zip(media_types, train_loaders))), skip_num=skip_num)

    model_without_ddp = model.module if config.distributed else model
    iterator = metric_logger.log_every(train_loader, log_freq, header)
    for i, (media_type, (image, text, idx)) in enumerate(iterator):
        image = image.to(device, non_blocking=True)
        idx = idx.to(device, non_blocking=True)
        text_input = tokenizer(text).to(device)

        with torch.cuda.amp.autocast(enabled=config.use_half_precision, dtype=data_type):
            loss_dict = model(image, text_input, idx=idx)
            loss = sum(loss_dict.values())
        
        if hasattr(config, "deepspeed") and config.deepspeed.enable:
            model.backward(loss)
            model.step()
        else: 
            if not config.use_half_precision or config.get('use_bf16', True):
                optimizer.zero_grad()
                loss.backward()
                if config.optimizer.max_grad_norm > 0:
                    torch.nn.utils.clip_grad_norm_(model.parameters(), config.optimizer.max_grad_norm)
                optimizer.step()
                scheduler.step()
            else:
                optimizer.zero_grad()
                scaler.scale(loss).backward()
                if config.optimizer.max_grad_norm > 0:
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(model.parameters(), config.optimizer.max_grad_norm)
                scaler.step(optimizer)
                scaler.update()
                scheduler.step()

        # logging
        for name in loss_names:
            value = loss_dict[name]
            value = value if isinstance(value, float) else value.item()
            metric_logger.update(**{f"{media_type}-{name}": value})
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])
        metric_logger.update(temperature=model_without_ddp.temp.item())

        if is_main_process() and config.wandb.enable and global_step % log_freq == 0:
            logs = metric_logger.get_global_avg_dict()
            log_dict_to_wandb(logs, step=global_step, prefix="train/")

        global_step += 1

        if config.debug and global_step % 20 == 0:
            logger.info("debug mode, break training loop")
            break

        if config.debug and global_step % (2 * log_freq + 3) == 0:
            logger.info("debug mode, break training loop")
            break

        if config.get('save_iter', 0) and global_step % config.save_iter == 0:
            if hasattr(config, "deepspeed") and config.deepspeed.enable:
                tag = f"ckpt_iter{global_step:02d}.pth"
                model.save_checkpoint(config.output_dir, tag=tag, save_latest=False, exclude_frozen_parameters=True)
            elif is_main_process():
                state_dict = model_without_ddp.state_dict()
                param_grad_dict = {
                    k: v.requires_grad for (k, v) in model_without_ddp.named_parameters()
                }
                for k in list(state_dict.keys()):
                    if k in param_grad_dict.keys() and not param_grad_dict[k]:
                        # delete parameters that do not require gradient
                        logger.info(f"Not saving {k}")
                        del state_dict[k]
                save_obj = {
                    "model": state_dict,
                    "optimizer": optimizer.state_dict(),
                    "scheduler": scheduler.state_dict(),
                    "scaler": scaler.state_dict(),
                    "config": config,
                    "epoch": epoch,
                    "global_step": global_step,
                }
                torch.save(save_obj, join(config.output_dir, f"ckpt_iter{global_step:02d}.pth"))

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    logger.info(f"Averaged stats: {metric_logger.global_avg()}")
    return global_step


def setup_dataloaders(config, mode="pt"):
    # train datasets, create a list of data loaders
    logger.info(f"Creating dataset for {mode}")
    train_datasets = create_dataset(f"{mode}_train", config)
    media_types = get_media_types(train_datasets)

    if config.distributed:
        batch_size = [config.inputs.batch_size[k] for k in media_types] # batch_size for each GPU
        samplers = create_stateful_sampler(train_datasets, batch_size)
    else:
        raise NotImplementedError

    train_loaders = create_loader(
        train_datasets,
        samplers,
        batch_size=[config.inputs.batch_size[k] for k in media_types],
        num_workers=[config.num_workers] * len(media_types),
        is_trains=[True] * len(media_types),
        collate_fns=[None] * len(media_types),
    )

    # test datasets, a mapping from dataset name to data loader
    test_datasets, test_dataset_names = create_dataset(f"{mode}_eval", config)
    test_loaders = create_loader(
        test_datasets,
        [None] * len(test_datasets),
        batch_size=[config.inputs.batch_size_test[d.media_type] for d in test_datasets],
        num_workers=[config.num_workers] * len(test_datasets),
        is_trains=[False] * len(test_datasets),
        collate_fns=[None] * len(test_datasets),
    )
    test_name2loaders = {k: v for k, v in zip(test_dataset_names, test_loaders)}
    return train_loaders, test_name2loaders, media_types


def main(config):
    if is_main_process() and config.wandb.enable:
        run = setup_wandb(config)

    is_pretrain = config.mode == "pt"

    logger.info(f"train_file: {config.train_file}")

    setup_seed(config.seed + get_rank())
    device = torch.device(config.device)

    train_loaders, test_name2loaders, train_media_types = setup_dataloaders(
        config, mode=config.mode
    )
    num_steps_per_epoch = sum(len(d) for d in train_loaders)

    config.scheduler.num_training_steps = num_steps_per_epoch * config.scheduler.epochs
    config.scheduler.num_warmup_steps = num_steps_per_epoch * config.scheduler.warmup_epochs
    # set cudnn.benchmark=True only when input size is fixed
    # https://discuss.pytorch.org/t/what-does-torch-backends-cudnn-benchmark-do/5936/3
    cudnn.benchmark = len(train_media_types) == 1

    model_cls = eval(config.model.get('model_cls', 'InternVideo2_CLIP'))
    (
        model,
        model_without_ddp,
        optimizer,
        scheduler,
        scaler,
        tokenizer,
        start_epoch,
        global_step,
    ) = setup_model(
        config,
        model_cls=model_cls,
        pretrain=is_pretrain,
        find_unused_parameters=True,
        num_steps_per_epoch=num_steps_per_epoch,
    )
    if is_main_process() and config.wandb.enable:
        wandb.watch(model)

    best = 0
    best_epoch = 0

    if config.get('use_bf16', True):
        data_type = torch.bfloat16
    else:
        data_type = torch.float16

    logger.info("Start training")
    logger.info(f"Epoch: {start_epoch}")
    start_time = time.time()
    start_step = start_epoch * num_steps_per_epoch
    for epoch in range(start_epoch, config.scheduler.epochs):
        if not config.evaluate:
            global_step = train(
                model,
                train_loaders,
                optimizer,
                tokenizer,
                epoch,
                global_step,
                device,
                scheduler,
                scaler,
                config,
                data_type,
                skip_num = global_step - start_step
            )

        # save checkpoint befor evaluation
        # only save those with gradient
        if hasattr(config, "deepspeed") and config.deepspeed.enable:
            if config.get("save_latest", False):
                tag = "ckpt_latest.pth"
            else:
                tag = f"ckpt_{epoch:02d}.pth"
            model.save_checkpoint(config.output_dir, tag=tag, save_latest=False, exclude_frozen_parameters=True)
            
        elif is_main_process():
            state_dict = model_without_ddp.state_dict()
            param_grad_dict = {
                k: v.requires_grad for (k, v) in model_without_ddp.named_parameters()
            }
            for k in list(state_dict.keys()):
                if k in param_grad_dict.keys() and not param_grad_dict[k]:
                    # delete parameters that do not require gradient
                    logger.info(f"Not saving {k}")
                    del state_dict[k]

            save_obj = {
                "model": state_dict,
                "optimizer": optimizer.state_dict(),
                "scheduler": scheduler.state_dict(),
                "scaler": scaler.state_dict(),
                "config": config,
                "epoch": epoch,
                "global_step": global_step,
            }
            if config.get("save_latest", False):
                torch.save(save_obj, join(config.output_dir, "ckpt_latest.pth"))
            else:
                torch.save(save_obj, join(config.output_dir, f"ckpt_{epoch:02d}.pth"))

        # evaluation
        with torch.cuda.amp.autocast(enabled=config.use_half_precision, dtype=data_type):
            eval_res = {}
            for test_name, test_loader in test_name2loaders.items():
                if test_name not in config.test_types:
                    logger.info(
                        f"Skip eval {test_name} split. All test_types {config.test_types}"
                    )
                    continue
                res = evaluation_wrapper(
                    model_without_ddp, test_loader, tokenizer, device, config, data_type=data_type, prefix=test_name
                )
                eval_res.update(res)

        # save the best checkpoint
        if is_main_process():
            # log to wandb
            if config.wandb.enable:
                for p, v in eval_res.items():
                    log_dict_to_wandb(v, step=global_step, prefix=p)

            if config.stop_key is not None and config.stop_key in eval_res:
                cur_r_mean = eval_res[config.stop_key]["r_mean"]
            else:  # None
                cur_r_mean = best + 1  # save the last as the best

            eval_res = pd.DataFrame(eval_res)
            logger.info(f"Epoch {epoch}")
            logger.info(f"\n{eval_res.transpose().to_string(max_cols=30)}")

            eval_res.to_json(join(config.output_dir, "eval_res_latest.json"))

            if not config.evaluate and cur_r_mean > best:
                if not hasattr(config, "deepspeed") or not config.deepspeed.enable:
                    torch.save(save_obj, join(config.output_dir, "ckpt_best.pth"))
                eval_file = "eval_res_best.json"
                eval_res.to_json(join(config.output_dir, eval_file))
                best = cur_r_mean
                best_epoch = epoch
        
        if hasattr(config, "deepspeed") and config.deepspeed.enable:
            r_mean_best = torch.tensor([0.0, 0.0]).to(device)
            if is_main_process():
                r_mean_best[0] = cur_r_mean
                r_mean_best[1] = best
            dist.broadcast(r_mean_best, 0)
            cur_r_mean, best = r_mean_best[0].item(), r_mean_best[1].item()
        
            if not config.evaluate and cur_r_mean > best:
                model.save_checkpoint(config.output_dir, tag="ckpt_best.pth", save_latest=False, exclude_frozen_parameters=True)

        if config.evaluate:
            break
        
        start_step = global_step

        dist.barrier()

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info(f"Training time {total_time_str}")
    logger.info(f"best epoch {best_epoch} [config.stop_key {config.stop_key}]")
    logger.info(f"Checkpoints and Logs saved at {config.output_dir}")

    if is_main_process() and config.wandb.enable:
        run.finish()


if __name__ == "__main__":
    cfg = setup_main()
    local_broadcast_process_authkey()
    main(cfg)



# --------------------------------------------------------
# BEATs: Audio Pre-Training with Acoustic Tokenizers (https://arxiv.org/abs/2212.09058)
# Github source: https://github.com/microsoft/unilm/tree/master/beats
# Copyright (c) 2022 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on VQGAN code bases
# https://github.com/CompVis/taming-transformers
# --------------------------------------------------------'

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as distributed

try:
    from einops import rearrange, repeat
except ImportError:
    pass


def l2norm(t):
    return F.normalize(t, p=2, dim=-1)


def ema_inplace(moving_avg, new, decay):
    moving_avg.data.mul_(decay).add_(new, alpha=(1 - decay))


def sample_vectors(samples, num):
    num_samples, device = samples.shape[0], samples.device

    if num_samples >= num:
        indices = torch.randperm(num_samples, device=device)[:num]
    else:
        indices = torch.randint(0, num_samples, (num,), device=device)

    return samples[indices]


def kmeans(samples, num_clusters, num_iters=10, use_cosine_sim=False):
    dim, dtype, device = samples.shape[-1], samples.dtype, samples.device

    means = sample_vectors(samples, num_clusters)

    for _ in range(num_iters):
        if use_cosine_sim:
            dists = samples @ means.t()
        else:
            diffs = rearrange(samples, 'n d -> n () d') \
                    - rearrange(means, 'c d -> () c d')
            dists = -(diffs ** 2).sum(dim=-1)

        buckets = dists.max(dim=-1).indices
        bins = torch.bincount(buckets, minlength=num_clusters)
        zero_mask = bins == 0
        bins_min_clamped = bins.masked_fill(zero_mask, 1)

        new_means = buckets.new_zeros(num_clusters, dim, dtype=dtype)
        new_means.scatter_add_(0, repeat(buckets, 'n -> n d', d=dim), samples)
        new_means = new_means / bins_min_clamped[..., None]

        if use_cosine_sim:
            new_means = l2norm(new_means)

        means = torch.where(zero_mask[..., None], means, new_means)

    return means, bins


class EmbeddingEMA(nn.Module):
    def __init__(self, num_tokens, codebook_dim, decay=0.99, eps=1e-5, kmeans_init=True, codebook_init_path=''):
        super().__init__()
        self.num_tokens = num_tokens
        self.codebook_dim = codebook_dim
        self.decay = decay
        self.eps = eps
        if codebook_init_path == '':
            if not kmeans_init:
                weight = torch.randn(num_tokens, codebook_dim)
                weight = l2norm(weight)
            else:
                weight = torch.zeros(num_tokens, codebook_dim)
            self.register_buffer('initted', torch.Tensor([not kmeans_init]))
        else:
            print(f"load init codebook weight from {codebook_init_path}")
            codebook_ckpt_weight = torch.load(codebook_init_path, map_location='cpu')
            weight = codebook_ckpt_weight.clone()
            self.register_buffer('initted', torch.Tensor([True]))

        self.weight = nn.Parameter(weight, requires_grad=False)
        self.cluster_size = nn.Parameter(torch.zeros(num_tokens), requires_grad=False)
        self.embed_avg = nn.Parameter(weight.clone(), requires_grad=False)
        # self.register_buffer('initted', torch.Tensor([not kmeans_init]))
        self.update = True

    @torch.jit.ignore
    def init_embed_(self, data):
        if self.initted:
            return
        print("Performing Kemans init for codebook")
        embed, cluster_size = kmeans(data, self.num_tokens, 10, use_cosine_sim=True)
        self.weight.data.copy_(embed)
        self.cluster_size.data.copy_(cluster_size)
        self.initted.data.copy_(torch.Tensor([True]))

    def forward(self, embed_id):
        return F.embedding(embed_id, self.weight)

    def cluster_size_ema_update(self, new_cluster_size):
        self.cluster_size.data.mul_(self.decay).add_(new_cluster_size, alpha=1 - self.decay)

    def embed_avg_ema_update(self, new_embed_avg):
        self.embed_avg.data.mul_(self.decay).add_(new_embed_avg, alpha=1 - self.decay)

    def weight_update(self, num_tokens):
        n = self.cluster_size.sum()
        smoothed_cluster_size = (
                (self.cluster_size + self.eps) / (n + num_tokens * self.eps) * n
        )
        # normalize embedding average with smoothed cluster size
        embed_normalized = self.embed_avg / smoothed_cluster_size.unsqueeze(1)
        # embed_normalized = l2norm(self.embed_avg / smoothed_cluster_size.unsqueeze(1))
        self.weight.data.copy_(embed_normalized)


def norm_ema_inplace(moving_avg, new, decay):
    moving_avg.data.mul_(decay).add_(new, alpha=(1 - decay))
    moving_avg.data.copy_(l2norm(moving_avg.data))


class NormEMAVectorQuantizer(nn.Module):
    def __init__(self, n_embed, embedding_dim, beta, decay=0.99, eps=1e-5,
                 statistic_code_usage=True, kmeans_init=False, codebook_init_path=''):
        super().__init__()
        self.codebook_dim = embedding_dim
        self.num_tokens = n_embed
        self.beta = beta
        self.decay = decay

        # learnable = True if orthogonal_reg_weight > 0 else False
        self.embedding = EmbeddingEMA(self.num_tokens, self.codebook_dim, decay, eps, kmeans_init, codebook_init_path)

        self.statistic_code_usage = statistic_code_usage
        if statistic_code_usage:
            self.register_buffer('cluster_size', torch.zeros(n_embed))
        if distributed.is_available() and distributed.is_initialized():
            print("ddp is enable, so use ddp_reduce to sync the statistic_code_usage for each gpu!")
            self.all_reduce_fn = distributed.all_reduce
        else:
            self.all_reduce_fn = nn.Identity()

    def reset_cluster_size(self, device):
        if self.statistic_code_usage:
            self.register_buffer('cluster_size', torch.zeros(self.num_tokens))
            self.cluster_size = self.cluster_size.to(device)

    def forward(self, z):
        # reshape z -> (batch, height, width, channel) and flatten
        # z, 'b c h w -> b h w c'
        # z = rearrange(z, 'b c h w -> b h w c')
        # z = z.transpose(1, 2)
        z = l2norm(z)
        z_flattened = z.reshape(-1, self.codebook_dim)

        self.embedding.init_embed_(z_flattened)

        d = z_flattened.pow(2).sum(dim=1, keepdim=True) + \
            self.embedding.weight.pow(2).sum(dim=1) - 2 * \
            torch.einsum('bd,nd->bn', z_flattened, self.embedding.weight)  # 'n d -> d n'

        encoding_indices = torch.argmin(d, dim=1)

        z_q = self.embedding(encoding_indices).view(z.shape)

        encodings = F.one_hot(encoding_indices, self.num_tokens).type(z.dtype)

        if not self.training:
            with torch.no_grad():
                cluster_size = encodings.sum(0)
                self.all_reduce_fn(cluster_size)
                ema_inplace(self.cluster_size, cluster_size, self.decay)

        if self.training and self.embedding.update:
            # EMA cluster size

            bins = encodings.sum(0)
            self.all_reduce_fn(bins)

            # self.embedding.cluster_size_ema_update(bins)
            ema_inplace(self.cluster_size, bins, self.decay)

            zero_mask = (bins == 0)
            bins = bins.masked_fill(zero_mask, 1.)

            embed_sum = z_flattened.t() @ encodings
            self.all_reduce_fn(embed_sum)

            embed_normalized = (embed_sum / bins.unsqueeze(0)).t()
            embed_normalized = l2norm(embed_normalized)

            embed_normalized = torch.where(zero_mask[..., None], self.embedding.weight,
                                           embed_normalized)
            norm_ema_inplace(self.embedding.weight, embed_normalized, self.decay)

        # compute loss for embedding
        loss = self.beta * F.mse_loss(z_q.detach(), z)

        # preserve gradients
        z_q = z + (z_q - z).detach()

        # reshape back to match original input shape
        # z_q, 'b h w c -> b c h w'
        # z_q = rearrange(z_q, 'b h w c -> b c h w')
        # z_q = z_q.transpose(1, 2)
        return z_q, loss, encoding_indices



"""Identity Loss & Face Recognition Score — reference-based face identity metric.

Measures how well a generated image/video preserves the identity of a reference
face.  Standard metric used by IP-Adapter, DreamBooth, InstantID, and other
identity-preserving generation pipelines.

Outputs:
    identity_loss          — cosine distance 0-1 (lower = better preservation)
    face_recognition_score — cosine similarity 0-1 (higher = better match)

Requires ``sample.reference_path`` pointing to a reference face image.
Gracefully skips when no reference is provided.

Tiered backends:
    1. InsightFace (buffalo_l ArcFace) — industry standard
    2. DeepFace (ArcFace) — fallback
    3. MediaPipe FaceMesh (geometric landmarks) — lightweight fallback
    4. Skip — no face models available
"""

import logging
from typing import Optional

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class IdentityLossModule(PipelineModule):
    name = "identity_loss"
    description = "Face identity preservation metric (cosine distance/similarity vs reference)"
    default_config = {
        "model_name": "buffalo_l",
        "subsample": 8,
        "warning_threshold": 0.5,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.model_name = self.config.get("model_name", "buffalo_l")
        self.subsample = self.config.get("subsample", 8)
        self.warning_threshold = self.config.get("warning_threshold", 0.5)
        self._backend = None  # "insightface" | "deepface" | "mediapipe"
        self._app = None  # InsightFace FaceAnalysis
        self._deepface = None
        self._mp_face_mesh = None

    def setup(self):
        # Tier 1: InsightFace
        try:
            from insightface.app import FaceAnalysis
            self._app = FaceAnalysis(name=self.model_name, providers=["CPUExecutionProvider"])
            self._app.prepare(ctx_id=-1, det_size=(640, 640))
            self._backend = "insightface"
            logger.info("IdentityLoss: using InsightFace backend.")
            return
        except Exception:
            pass

        # Tier 2: DeepFace
        try:
            from deepface import DeepFace
            self._deepface = DeepFace
            self._backend = "deepface"
            logger.info("IdentityLoss: using DeepFace backend.")
            return
        except Exception:
            pass

        # Tier 3: MediaPipe FaceMesh
        try:
            import mediapipe as mp
            self._mp_face_mesh = mp.solutions.face_mesh.FaceMesh(
                static_image_mode=True, max_num_faces=1, min_detection_confidence=0.5
            )
            self._backend = "mediapipe"
            logger.info("IdentityLoss: using MediaPipe landmark fallback.")
            return
        except Exception:
            pass

        logger.warning("IdentityLoss: no face backend available — module disabled.")

    def process(self, sample: Sample) -> Sample:
        if self._backend is None:
            return sample

        ref_path = getattr(sample, "reference_path", None)
        if ref_path is None:
            return sample

        try:
            ref_img = cv2.imread(str(ref_path))
            if ref_img is None:
                return sample
            ref_rgb = cv2.cvtColor(ref_img, cv2.COLOR_BGR2RGB)

            frames = self._load_frames(sample)
            if not frames:
                return sample

            if self._backend == "insightface":
                distance = self._compute_insightface(ref_rgb, frames)
            elif self._backend == "deepface":
                distance = self._compute_deepface(ref_path, sample, frames)
            else:
                distance = self._compute_mediapipe(ref_rgb, frames)

            if distance is None:
                return sample

            distance = float(np.clip(distance, 0.0, 1.0))
            similarity = 1.0 - distance

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.identity_loss = distance
            sample.quality_metrics.face_recognition_score = similarity

        except Exception as e:
            logger.warning(f"IdentityLoss failed for {sample.path}: {e}")

        return sample

    # -- Backend implementations ------------------------------------------------

    def _compute_insightface(self, ref_rgb, frames):
        ref_faces = self._app.get(ref_rgb)
        if not ref_faces:
            return None
        ref_emb = ref_faces[0].embedding
        ref_emb = ref_emb / (np.linalg.norm(ref_emb) + 1e-10)

        distances = []
        for frame in frames:
            faces = self._app.get(frame)
            if not faces:
                continue
            emb = faces[0].embedding
            emb = emb / (np.linalg.norm(emb) + 1e-10)
            cos_sim = float(np.dot(ref_emb, emb))
            distances.append(1.0 - cos_sim)

        return float(np.mean(distances)) if distances else None

    def _compute_deepface(self, ref_path, sample, frames):
        import tempfile
        import os
        from PIL import Image

        distances = []
        for frame in frames:
            with tempfile.NamedTemporaryFile(suffix=".jpg", delete=False) as tmp:
                tmp_path = tmp.name
                Image.fromarray(frame).save(tmp_path)
            try:
                result = self._deepface.verify(
                    img1_path=str(ref_path),
                    img2_path=tmp_path,
                    model_name="ArcFace",
                    enforce_detection=False,
                )
                distances.append(result["distance"])
            except Exception:
                pass
            finally:
                os.unlink(tmp_path)

        return float(np.mean(distances)) if distances else None

    def _compute_mediapipe(self, ref_rgb, frames):
        ref_lm = self._extract_landmarks(ref_rgb)
        if ref_lm is None:
            return None

        distances = []
        for frame in frames:
            lm = self._extract_landmarks(frame)
            if lm is None:
                continue
            dist = float(np.mean(np.linalg.norm(ref_lm - lm, axis=1)))
            distances.append(min(dist, 1.0))

        return float(np.mean(distances)) if distances else None

    def _extract_landmarks(self, rgb_image):
        """Extract normalized 468-point face landmarks from an RGB image."""
        results = self._mp_face_mesh.process(rgb_image)
        if not results.multi_face_landmarks:
            return None
        lm = results.multi_face_landmarks[0]
        pts = np.array([[p.x, p.y, p.z] for p in lm.landmark])
        centroid = pts.mean(axis=0)
        pts = pts - centroid
        scale = np.linalg.norm(pts, axis=1).max() + 1e-10
        pts = pts / scale
        return pts

    # -- Frame loading ----------------------------------------------------------

    def _load_frames(self, sample: Sample):
        frames = []
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    cap.release()
                    return frames
                n = min(self.subsample, total)
                indices = np.linspace(0, total - 1, n, dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
                cap.release()
            else:
                img = cv2.imread(str(sample.path))
                if img is not None:
                    frames.append(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        except Exception as e:
            logger.debug(f"Frame loading failed: {e}")
        return frames



import logging
import numpy as np
import torch
import torch.nn.functional as F
from peft import get_peft_model, LoraConfig, TaskType
from torch import nn
from transformers import LlamaForCausalLM, LlamaConfig

from transformers import LlamaTokenizer

logger = logging.getLogger(__name__)


class LLaMA(nn.Module):
    def __init__(
            self,
            use_flash_attn: bool = True,
            transformer_width: int = 4096,
            llama_path: str = None,
            use_lora: bool = True,
            clip_embed_dim: int = 768,
        ):
        super().__init__()
        
        self.use_flash_attn = use_flash_attn
        self.transformer_width = transformer_width
        
        """ text encoder of InternVL """
        llama_config = LlamaConfig.from_pretrained(llama_path, local_files_only=True)
        llama_config.causal = True
        llama_config.use_flash_attention = use_flash_attn
        # model = LlamaForCausalLM.from_pretrained(  # LLAMA model
        #     llama_path, torch_dtype=torch.float16, config=llama_config, local_files_only=True)
        model = LlamaForCausalLM(config=llama_config)
        if not use_lora:
            self.transformer = model.model
        else:
            peft_config = LoraConfig(
                task_type=TaskType.CAUSAL_LM, inference_mode=False, r=16, lora_alpha=32, lora_dropout=0.1)
            model = get_peft_model(model, peft_config)
            self.transformer = model.base_model.model.model
        
        self.transformer.gradient_checkpointing = True
        self.text_projection = nn.Parameter(torch.empty(transformer_width, clip_embed_dim))
            
    def forward(self, text):
        text_key_padding_mask = text > 0
        
        x = self.transformer(input_ids=text, attention_mask=text_key_padding_mask).last_hidden_state
        x = x[torch.arange(x.shape[0]), text_key_padding_mask.sum(1) - 1]
        x = x @ self.text_projection

        return x


class Tokenizer(nn.Module):
    def __init__(self, tokenizer_path="your_model_path/chinese_alpaca_lora_7b"):
        super(Tokenizer, self).__init__()
        self.tokenizer = LlamaTokenizer.from_pretrained(
            tokenizer_path, 
            local_files_only=True,
            legacy=False
        )
        self.tokenizer.pad_token = " "  # allow padding
        self.tokenizer.add_eos_token = True
    
    def forward(self, text):
        text = ["summarize:" + item for item in text]
        text = self.tokenizer(text, return_tensors="pt", max_length=80, truncation=True, padding="max_length").input_ids
        return text
